#!/usr/bin/env python

"""
Module containing publishing conversions

This module contains methods which convert topic, payload, config and sensors meta data into different MQTT
publishing formats.

The following methods are contained:
|     method      |  version |  tested  |              comment             | manual | *used by |
| --------------- |  ------- |  ------- |  ------------------------------- | ------ | -------- |
|  martas         |    2.0.0 |      yes |                                  | -       | lib     |
|  intermagnet    |    2.0.0 |       -  |                                  | -       | lib     |

"""

import unittest
import os
import sys
from datetime import datetime
from json import JSONEncoder, dumps
import numpy as np
from magpy.core.methods import is_number


def martas(pubdict=None, topic="", data="", head="", count=0, changecount=10, imo="TST", meta=None):
    """
    DESCRIPTION
        martas publishing format will create a CSV like output and sends meta information in a reduced cadence
    APPLICATION
        pubdict, count = martas(None, topic=topic, data=data, head=head, count=self.count, changecount=self.metacnt,
                imo="self.confdict.get('station', ''), meta=self.sensordict)
        self.count = count
    """
    if not meta:
        meta = {}
    if not pubdict:
        pubdict = {}

    pubdict[topic + "/data"] = data
    if count == 0:
        ## 'Add' is a string containing dict info like:
        ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
        add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format(
            meta.get('sensorid', ''), imo, meta.get('pierid', ''), meta.get('protocol', ''), meta.get('sensorgroup', ''),
            meta.get('sensordesc', ''), meta.get('ptime', ''))

        pubdict[topic + "/dict"] = add
        pubdict[topic + "/meta"] = head
    count += 1
    if count >= changecount:
        count = 0
    return pubdict, count


def intermagnet(pubdict=None, topic="", data="", head="", imo="TST", meta=None):
    """
    DESCRIPTION
        intermagnet publishing format will create a json style output as defined here

        # Topic
        # impf/<iaga-code>/<cadence>/<publication-level>/<elements-recorded>
        # impf/esk/pt1m/1/hdzs
        # Payload
        # {
        #     "startDate": "2023-01-01T00:00",
        #     "geomagneticFieldX": [ 17595.02, null, 17594.99 ],
        #     "geomagneticFieldY": [ -329.19, -329.18, -329.21 ],
        #     "geomagneticFieldZ": [ 46702.70, 46703.01, 46703.24 ]
        # }
        # "ginCode": (IMF)
        # "decbas": (IMF)
        # "latitude": (IMF, IAGA-2002, ImagCDF)
        # "longitude": (IMF, IAGA-2002, ImagCDF)
        # "elevation": (IAGA-2002, ImagCDF)
        # "institute": (IAGA-2002, ImagCDF - called "Source of data" in IAGA-2002)
        # "name": (IAGA-2002, ImagCDF - called "ObservatoryName" in ImagCDF)
        # "sensorOrientation": (IAGA-2002, ImagCDF - called "VectorSensOrient in CDF)
        # "digitalSampling": (IAGA-2002)
        # "dataIntervalType": (IAGA-2002)
        # "publicationDate": (IAGA-2002, ImagCDF)
        # "standardLevel": (ImagCDF)
        # "standardName": (ImagCDF)
        # "standardVersion": (ImagCDF)
        # "partialStandDesc": (ImagCDF)
        # "source": (ImagCDF)
        # "termsOfUse": (ImagCDF)
        # "uniqueIdentifier": (ImagCDF)
        # "parentIdentifiers": (ImagCDF)
        # "referenceLinks": (ImagCDF)
        # "comments": (IAGA-2002)
    APPLICATION
        pubdict = intermagnet(None, topic=topic, data=data, head=head, count=self.count, changecount=self.metacnt,
                imo="self.confdict.get('station', ''), meta=self.sensordict)
        self.count = count
    """
    debug = True
    datelist = []
    datablock = {}
    if not pubdict:
        pubdict = {}

    def _get_cadence(samplingperiod):
        if 0.9 < samplingperiod < 1.1:
            cadence = "pt1s"
            dtformat = "%Y-%m-%dT%H:%M:%S"
        elif 58 < samplingperiod < 62:
            cadence = "pt1m"
            dtformat = "%Y-%m-%dT%H:%M"
        elif 3500 < samplingperiod < 3700:
            cadence = "pt1h"
            dtformat = "%Y-%m-%dT%H"
        elif 86000 < samplingperiod < 87000:
            cadence = "pt1d"
            dtformat = "%Y-%m-%d"
        else:
            print ("Samplingperiod {} not supported by INTERMAGNET MQTT payload format")
            cadence = "{:.1f}S".format(samplingperiod)
            dtformat = "%Y-%m-%dT%H:%M:%S"
        return cadence, dtformat

    def _get_components(head,meta):
        components = None
        component1 = meta.get("SensorElements", None)
        component2 = meta.get("DataComponents", None)
        try:
            component3 = head.split()[4].strip("[").strip("]").replace(",", "")
        except:
            component3 = None
        if not components:
            components = component3
        if not components:
            components = component2
        if not components:
            components = component1
        return components.lower()

    def nan_to_none(obj):
        if isinstance(obj, dict):
            return {k: nan_to_none(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [nan_to_none(v) for v in obj]
        elif isinstance(obj, float) and np.isnan(obj):
            return None
        return obj

    class NanConverter(JSONEncoder):
        def encode(self, obj, *args, **kwargs):
            return super().encode(nan_to_none(obj), *args, **kwargs)

    # get the following info from file header
    if debug:
        print ("HEADER", head, meta)
    multilist = list(map(float, head.split()[6].strip('[').strip(']').split(',')))
    samplingperiod = float(meta.get("DataSamplingRate",1.0))
    publevel = meta.get("DataPublicationLevel", 1)
    components = _get_components(head,meta)
    # fill in header info
    if meta.get("ginCode", None):
        datablock["ginCode"] = meta.get("ginCode")
    if meta.get("decbas", None):
        datablock["decbas"] = meta.get("decbas")
    # "latitude": (IMF, IAGA-2002, ImagCDF)
    if meta.get("DataAcquisitionLatitude", None):
        if is_number(meta.get("DataAcquisitionLatitude")):
            datablock["latitude"] = float(meta.get("DataAcquisitionLatitude"))
        else:
            datablock["latitude"] = meta.get("DataAcquisitionLatitude")
    # "longitude": (IMF, IAGA-2002, ImagCDF)
    if meta.get("DataAcquisitionLongitude", None):
        if is_number(meta.get("DataAcquisitionLongitude")):
            datablock["longitude"] = float(meta.get("DataAcquisitionLongitude"))
        else:
            datablock["longitude"] = meta.get("DataAcquisitionLongitude")
    # "elevation": (IAGA-2002, ImagCDF)
    if meta.get("DataElevation", None):
        if is_number(meta.get("DataElevation")):
            datablock["elevation"] = float(meta.get("DataElevation"))
        else:
            datablock["elevation"] = meta.get("DataElevation")
    if meta.get('StationInstitution'):
        datablock["institute"] = meta.get("StationInstitution")
    # "institute": (IAGA-2002, ImagCDF - called "Source of data" in IAGA-2002)
    if meta.get('StationName'):
        datablock["name"] = meta.get("StationName")
    # "name": (IAGA-2002, ImagCDF - called "ObservatoryName" in ImagCDF)
    if meta.get('DataSensorOrientation'):
        datablock["sensorOrientation"] = meta.get("DataSensorOrientation")
    # "sensorOrientation": (IAGA-2002, ImagCDF - called "VectorSensOrient in CDF)
    if meta.get('DataDigitalSampling'):
        datablock["digitalSampling"] = meta.get("DataDigitalSampling")
    # "digitalSampling": (IAGA-2002)
    if meta.get('DataSamplingFilter'):
        datablock["dataIntervalType"] = meta.get("DataSamplingFilter")
    # "dataIntervalType": (IAGA-2002)
    if meta.get('DataPublicationDate'):
        datablock["publicationDate"] = str(meta.get("DataPublicationDate"))
    # "publicationDate": (IAGA-2002, ImagCDF)
    if meta.get('DataStandardLevel'):
        datablock["standardLevel"] = meta.get("DataStandardLevel")
    # "standardLevel": (ImagCDF)
    if meta.get('DataStandardName'):
        datablock["standardName"] = meta.get("DataStandardName")
    # "standardName": (ImagCDF)
    if meta.get('DataStandardVersion'):
        datablock["standardVersion"] = meta.get("DataStandardVersion")
    # "standardVersion": (ImagCDF)
    if meta.get('DataPartialStandDesc'):
        datablock["partialStandDesc"] = meta.get("DataPartialStandDesc")
    # "partialStandDesc": (ImagCDF)
    if meta.get('DataSource'):
        datablock["source"] = meta.get("DataSource")
    # "source": (ImagCDF)
    if meta.get('DataTerms'):
        datablock["termsOfUse"] = meta.get("DataTerms")
    # "termsOfUse": (ImagCDF)
    if meta.get('DataID'):
        datablock["uniqueIdentifier"] = meta.get("DataID")
    # "uniqueIdentifier": (ImagCDF)
    if meta.get('SensorID'):
        datablock["ParentIdentifiers"] = meta.get("SensorID")
    # "parentIdentifiers": (ImagCDF)
    if meta.get('StationWebInfo'):
        datablock["referenceLinks"] = meta.get("StationWebInfo")
    # "referenceLinks": (ImagCDF)
    if meta.get('DataComments'):
        datablock["comments"] = meta.get("DataComments")
    # "comments": (IAGA-2002)
    #if meta.get('StationK9'):
    #    datablock["k9level"] = meta.get("StationK9")

    lines = data.split(";")
    # Get the startdate
    dateext = lines[0].split(",")[:7]
    dateline = list(map(int, dateext))
    startdt = datetime(*dateline)

    for line in lines:
        data = line.split(",")
        # times and data
        if len(data)>7:
            dt = data[:7]
            dtn = list(map(int, dt))
            datelist.append(datetime(*dtn))
            #add all dates to a list
            dat = data[7:]
            if debug:
                print ("DATA", dat)
                print ("LEN", len(dat))
                print ("LEN Comp", len(components))
            if len(dat) >= len(components):
                pass
            else:
                components = components[:len(dat)]
            # Limit the amount of components to 4
            components = components[:4]
            for idx,el in enumerate(components):
                mu = 1.0
                if len(multilist) >= idx:
                   mu = multilist[idx]
                   if not is_number(mu):
                       mu = 1.0
                   mu = float(mu)
                val = dat[idx]
                blockname = f"geomagneticField{el.upper()}"
                if debug:
                    print(blockname)
                l = datablock.get(blockname, [])
                if is_number(val) and np.isnan(float(val)):
                    val = None
                l.append(float(val)/mu)
                datablock[blockname] = l
        else:
            # no data
            pass
    # determine increment of date list and eventually update cadence
    if len(datelist) > 1 and not samplingperiod:
        meand = np.mean(np.diff(datelist))
        samplingperiod = meand.total_seconds()
    cadence, dtformat = _get_cadence(samplingperiod)
    starttime = startdt.strftime(dtformat)
    datablock["startDate"] = starttime

    topic = "impf/{}/{}/{}/{}".format(imo.lower(), cadence, publevel, components)

    # replace NaN with null - Test set
    #datablock["geomagneticFieldX"] = [ 17595.02, np.nan, 17594.99 ]
    #datablock["geomagneticFieldY"] = [ -329.19, -329.18, -329.21 ]
    #datablock["geomagneticFieldZ"] = [ 46702.70, 46703.01, 46703.24 ]

    # convert datablock to json
    datajson = dumps(datablock, cls=NanConverter)
    pubdict[topic] = datajson

    return pubdict


#!/usr/bin/env python
# coding=utf-8
"""

DESCRIPTION
    Filtering data files. By default all results are written to table '0002'. A DATAINFO input is not created. To do that add option -s and specify the sensor(s).


PARAMETER:
    sensors:

APPLICATION
    filter.py -j archive for daily jobs (read HF cdf archive files)
    filter.py -j realtime for permanent jobs (read HF database data)

filter.py -c singlefilter.cfg -j archive,second -d 5 -e 2021-02-01


# Better read a json stucture
a

filter.py requires a file called realtime.py within the same directory. realtime.py looks like:


filter.cfg

{"filter": {
              "groupparameterdict":{ "LEMI" : {"station" : "WIC", "filtertype":"default", "realtime" : "True", "window" : 40000},
                                     "LEMI036_3_0001_0001" : {"filtertype":"hann", "filterwidth":100, "resample_period":1, "realtime":"False", "window" : 40000},
                                     "LEMI025_28_0002_0001" : {"filtertype":"default", "realtime":"False", "window" : 40000},
                                     "GSM90_14245_0002_0001" : {"filtertype":"gaussian", "filterwidth":3.33333333, "resample_period":1, "realtime":"True", "window" : 10000},
                                     "IWT_TILT01_0001_0003" : {"filtertype": "default", "realtime" : "False", "window" : 40000},
                                     "BM35" : {"filtertype":"gaussian", "filterwidth":3.33333333, "resample_period":1, "realtime":"True", "window" : 10000},
                                     "G823A" : {"filtertype": "default", "realtime" : "False", "window" : 40000}
                                    },
              "scalardict":{"GP20S3NS_012201_0001_0001": {"type" : "GP20S3", "components" : ["x","y","z"]},
                             "GP20S3V_911005_0001_0001" :  {"type" : "GP20S3", "components" : ["x","y","z"]},
                             "GP20S3EW_111201_0001_0001" :  {"type" : "GP20S3", "components" : ["x","y","z"]}
                                    },
              "realtime": ["LEMI025_22_0003_0001","LEMI036_1_0002_0001","LEMI036_3_0001_0001","GSM90_14245_0002_0001","BM35_029_0001_0001","BM35_033_0001_0001"],
              "blacklist": ["LEMI025_28_0002_0001"],
              "basics": {"basepath":"/srv/archive","outputformat":"PYCDF","destination":"db"}
            }
}


# If you just want to filter data for archiving directly to disk only of a single sensor use
singlefilter.cfg
{"filter": {
              "groupparameterdict":{ "LEMI025_28_0002_0001" : {"filtertype":"default", "realtime":"False", "window" : 40000}},
              "scalardict":{},
              "realtime": [],
              "blacklist": [],
              "basics": {"basepath":"/srv/archive","outputformat":"PYCDF","destination":"disk"}
            }
}

and use:
filter.py -c singlefilter.cfg -j archive,second -d 5 -e 2021-02-01

Please note that filtering always reads the complete timerange..thus for long time ranges construct a bash job like:
filter.py -c singlefilter.cfg -j archive,second -d 2 -e 2021-02-01
filter.py -c singlefilter.cfg -j archive,second -d 2 -e 2021-01-30
filter.py -c singlefilter.cfg -j archive,second -d 2 -e 2021-02-28
filter.py -c singlefilter.cfg -j archive,second -d 2 -e 2021-02-26


| class  |    method     | since vers |  validation   |  comment    | manual  |  *used by |
| ------ |  -----------  | ---------- | ------------- | ----------  | ------- | --------- |
|        |               |            |               |             |         |           |
|        |  read_conf    |      2.0.0 |  yes          |             |         |           |
|        |  get_sensors  |      2.0.0 |  yes          |             |         |           |


"""

from magpy.stream import *
from magpy.core import database
from magpy.core import methods
import magpy.opt.cred as mpcred

import getopt
import socket
import json

from martas.version import __version__
from martas.core.methods import martaslog as ml
from martas.core import methods as mm


def read_conf(path):
    """
    DESCRIPTION
    Read a dictionary structure from a text file
    """
    filterdict = {}
    if os.path.isfile(path):
        with open(path, "r") as f:
             contents = f.read()
             c = contents.replace(" ","").replace("\n","")
             d = json.loads(c)
             filterdict = d.get('filter')
    return filterdict



def get_sensors(db=None,groupdict=None,samprate=None,recent=False,recentthreshold=7200,blacklist=None, debug=False):
    """
    DESCRIPTION
      extracts Sensors with names similar a the one in groupdict
      You can also select for sampling rates and recent data

    REQUIREMENTS
      an input in DATAINFO is required
        -> check G823 on Aldebaran
        -> highest res always to 0001 -> why are there so many LEMI028 inputs in table
    """

    if not groupdict:
        groupdict = {}
    if not blacklist:
        blacklist = []
    returndict = {}
    timerange = recentthreshold # used for sampling rate determination
    dbdateformat = "%Y-%m-%d %H:%M:%S.%f"

    for key in groupdict:
        print ("Get-senors: Dealing with group", key)
        # Checking all available sensors
        # ##############################
        samprate = groupdict[key].get('samplingrate',samprate)
        srlist = []
        stationid = groupdict[key].get('station',None)
        stationstr = ''
        if stationid:
            stationstr = ' AND StationID LIKE "{}"'.format(stationid)
        sensorlist = db.select('DataID', 'DATAINFO','DataID LIKE "{}%"{}'.format(key,stationstr))
        if debug:
            print ("   -> Found {}".format(sensorlist))

        # Apply black list
        # ##############################
        sensorlist = [sens for sens in sensorlist if not sens in blacklist]

        # If sampling rate criteria is given then subselect all sensors
        # ##############################
        if samprate:
            if debug:
                print ("- Limiting selection to sensors with sampling rate {}".format(samprate))
            validsensors1 = []  # add sensors for which the sampling rate is given in DATAINFO
            determinesr = [] # add senors for which the sampling rate needs to be determined
            srlist = []
            if debug:
                print (" * Selecting sensors with appropriate sampling rate in DATAINFO")
            for sensor in sensorlist:
                res = db.select('DataSamplingrate','DATAINFO','DataID="{}"'.format(sensor))
                try:
                    sr = float(res[0])
                    if debug:
                        print ("    Sensor (DATAINFO): {} -> Samplingrate: {}".format(sensor,sr))
                    if samprate == 'HF':
                        if sr < 0.98:
                            validsensors1.append(sensor)
                            srlist.append(sr)
                    else:
                        try:
                            samprate = float(samprate)
                        except:
                            samprate = 1.0
                        if samprate*0.98 <= sr and sr <= samprate*1.02:
                            validsensors1.append(sensor)
                            srlist.append(sr)
                except:
                    print ("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                    print ("Check sampling rate {} of {}".format(res,sensor))
                    print ("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                    determinesr.append(sensor)

            if len(determinesr) > 0 and debug:
                print (" * checking sampling rate of sensors with no sampling rate given in DATAINFO")
            for sensor in determinesr:
                lastdata = db.get_lines(sensor,timerange)
                if lastdata.length()[0] > 0:
                    sr = lastdata.samplingrate()
                    if debug:
                        print ("    Sensor (TESTING DATA): {} -> Samplingrate: {}".format(sensor,sr))
                    if samprate == 'HF':
                        if sr < 0.98:
                            validsensors1.append(sensor)
                            srlist.append(sr)
                    else:
                        try:
                            samprate = float(samprate)
                        except:
                            samprate = 1.0
                        if samprate*0.98 <= sr and sr <= samprate*1.02:
                            validsensors1.append(sensor)
                            srlist.append(sr)
                    # update data of these sensors in db
                    if debug:
                        print("new header of sensor without samplingrate:", lastdata.header)
                    if db and not debug:
                        # This will only be done once for a new sensor without sampling rate in DATAINFO
                        print("updating header with determined sampling rate:", lastdata.header)
                        db.write(lastdata)
            sensorlist = validsensors1
            if debug:
                print (" -> got {} sensors".format(len(sensorlist)))
        # If only sensors with current data should be used
        # ##############################
        if recent:
            currentthreshold = datetime.now(timezone.utc).replace(tzinfo=None) - timedelta(seconds=recentthreshold)
            if debug:
                print ("- Limiting selection to sensors with current data")
            validsensors = []
            validsr = []
            for idx,sensor in enumerate(sensorlist):
                last = db.select('time',sensor,expert="ORDER BY time DESC LIMIT 10")
                if len(last) > 0:
                    dbdate = last[0]
                    if dbdate > currentthreshold:
                        if debug:
                            print ("   current data for {}".format(sensor))
                        validsensors.append(sensor)
                        validsr.append(srlist[idx])
            sensorlist = validsensors

        if debug:
            print(" -> finally got {} sensors for filtering".format(len(sensorlist)))
        # join all obtained sensors into a new dictionary
        # ##############################
        for sensor in sensorlist:
            returndict[sensor] = groupdict.get(key)

    return returndict


def apply_filter(db, statusmsg=None, groupdict=None, permanent=None, blacklist=None, jobtype='realtime', endtime=datetime.now(timezone.utc).replace(tzinfo=None), dayrange=2, basepath='', dbinputsensors=None, destination='db', outputformat='PYCDF', recentthreshold=7200, debug=False):
    """
    DESCRIPTION
        Create one-second records by filtering data sets

    PARAMETER
        jobtype  : 'realtime' or 'archive'
        destination : 'db' or 'disk' -> if db (for realtime), disk archiving is done by archive; disk should only be used to convert old files 
    """
    # Allow to define groups in highreslist (e.g. LEMI036*)
    # Then firstly extract all instruments which fit to the appropriate group
    # and assign the parameters to each instrument.
    # Later list positions need to overwrite previous inputs: e.g.
    #   LEMI036*, par1, par2
    #   LEMI036_3, par4, par5
    if not statusmsg:
        statusmsg = {}
    if not groupdict:
        groupdict = {}
    if not permanent:
        permanent = []
    if not blacklist:
        blacklist = []
    if not dbinputsensors:
        dbinputsensors = []
    stationid = ''
    sensor = ''

    if debug:
        print ("Selected options are: Destination = {}; outputformat = {}; path = {}".format(destination, outputformat, basepath))

    dayrangedefault = dayrange
    sn = socket.gethostname()
    recent = True
    if jobtype == 'archive':
        recent = False

    highreslst = get_sensors(db=db,groupdict=groupdict,samprate='HF',blacklist=blacklist,recent=recent,recentthreshold=recentthreshold,debug=debug)
    if debug:
        print ("  Obtained HF list (sampling rate higher then 1 Hz):", highreslst)
        print ("  Starting one second data filtering:  (Filtering high resolution data sets)")
    p1start = datetime.now()
    for inst in highreslst:
        if debug:
            print ("  ----------------------------")
            print ("  Dealing with instrument {}".format(inst))
        name = '{}-filter-{}-{}'.format(sn,jobtype,inst.replace('_',''))
        last = DataStream()
        dataexpected = True
        try:
            noresample = None
            resample_period = None
            options = highreslst[inst]
            if debug:
                print ("     -> Selected options:", options)
                print ("  Obtaining projected data amount")
            if jobtype == 'realtime' and inst in permanent:
                if debug:
                    print ("     Realtime job selected -> getting data from database")
                amount = options.get('window',40000)
                rt = options.get('realtime',True)
                # amount * sampling rate defines coverage in seconds -> 10Hz (0.1 sec * 40000 -> 4000 sec or 0.5 sec * 10000 -> 5000 sec)
                if not mm.get_bool(rt):
                    amount = amount*3                # triple the amount of expected data if only hourly uploads are existing
                last = db.get_lines(inst,amount)
                try:
                    test = len(last)
                except:
                    statusmsg[name] = 'db access - data not existing'
                if debug:
                    print ("     -> Extracting {} data points".format(amount))
                if len(last) > 0:
                    sr = last.samplingrate()
                    expectedcoverage_in_sec = amount*sr
                    now = datetime.now(timezone.utc).replace(tzinfo=None)
                    last = last.trim(starttime=now-timedelta(seconds=expectedcoverage_in_sec),endtime=now) # remove all timesteps exceeding current time (typical IWT error)
                    if debug:
                        print ("     -> Extracted {} data points in reality after slicing".format(len(last)))
                        print ("     -> Done")
            elif jobtype == 'realtime' and not inst in permanent:
                dataexpected = False
            elif not jobtype == 'realtime':
                dayrange = options.get('dayrange',dayrangedefault)
                if debug:
                    print ("     Archive job selected -> getting data from archive")
                stationid = db.select('StationID', 'DATAINFO', 'DataID LIKE "{}"'.format(inst))[0]
                sensor = "_".join(inst.split('_')[:-1])
                datapath = os.path.join(basepath,stationid,sensor,inst,'*')
                if debug:
                    print ("Datapath", datapath)
                begin = endtime-timedelta(days=dayrange)
                if debug:
                    print ("    Reading data between {} and {}".format(begin, endtime))
                last = read(datapath,starttime=begin,endtime=endtime)
                if debug:
                    print ("     -> found data between {}".format(last.timerange()))
                    print ("         corresponding to {} datapoints".format(len(last)))
                    print ("     -> Done")

            if debug:
                print ("    Got {} datapoints".format(last.length()[0]))

            if last.length()[0] > 0:
                if options.get('filtertype') == 'default':
                    filtertype = 'gaussian'
                    filterwidth = None # use default of 3.3333333 seconds
                else:
                    filtertype = options.get('filtertype','gaussian')
                    fwin = options.get('filterwidth',None)
                    filterwidth = timedelta(seconds=fwin)
                    resamp = options.get('resample_period',None)
                    if resamp == 'noresample':
                        noresample = True
                    else:
                        try:
                            resample_period=int(resamp)
                        except:
                            resample_period=1.0
                            pass
                missingdata = options.get('missingdata', 'conservative')
                destrevision = options.get('revision', '0002')
                if debug:
                    print ("    Default analysis parameter (may be re-specified for individual sensors): {}, filter type: {}, filter width: {}, resample period: {}, {}".format(inst, filtertype, filterwidth, resample_period, noresample))
                filtstream = last.filter(filter_type=filtertype, filter_width=filterwidth, missingdata=missingdata,resample_period=resample_period,noresample=noresample)
                # cut out the last 90% to reduce boundary filter effects # after 0.4.6
                try:
                    if debug:
                        print ("    Cutting out two lines (seconds) from start and two from beginning ..")
                    am = len(filtstream)
                    print (len(filtstream))
                    filtstream = filtstream.cut(am-2, 1, 0)
                    filtstream = filtstream.cut(am-4, 1, 1)
                    print ("Should be length -4", len(filtstream))
                    if debug:
                        print ("     Filtstream coverage: length={}, {}".format(len(filtstream), filtstream.timerange()) )
                        print ("     -> Done")
                except:
                    pass

                #### ALWAYS write to 0002 table
                newtab = "{}_{}".format(inst[:-5],destrevision)

                if not destination == 'disk':
                    # Write to database
                    try:
                        if not debug:
                            if newtab in dbinputsensors:
                                # if the sensor is already contained in DATAINFO then solely write data contents
                                db.write(filtstream,tablename=newtab)
                            else:
                                #print ("   Sensor contained in DBlist - adding Metainformation to DATAINFO")
                                db.write(filtstream)
                            #print ("    Writing to DB successful")
                        else:
                            print ("    !! Debug selected - skipping writing to DB")
                    except:
                        """db.write errors currently not captured - 
                           only general DB connection failures.
                           it happens once in a while that DB connection
                           is lost before writing (on sol at 3:00 and 4:00 UTC)
                           eventually connected to other DB access e.g.di analyses
                        """
                        pass
                else:
                    if debug:
                        print ("   Writing data directly to disk...")
                    archivepath = os.path.join(basepath,stationid,sensor,newtab)
                    if debug:
                        print ("   Destinationpath: {}".format(archivepath))
                    if not debug and outputformat and archivepath:
                        #print ("     valid write conditions")
                        filtstream.write(archivepath,filenamebegins=newtab+'_',format_type=outputformat)
                        #print ("    -> Done")
                    else:
                        print ("   Debug: skip writing")
                        print ("    -> without debug a file with {} inputs would be written to {}".format(filtstream.length()[0],archivepath))

                statusmsg[name] = 'fine'
            else:
                # send no data monitor
                if dataexpected:
                    statusmsg[name] = 'no data found'
        except:
            # send failed message to monitor
            statusmsg[name] = 'one second filter: general failure'

    if debug:
        p1end = datetime.now()
        print ("   One second job needed {}".format(p1end-p1start))
    return statusmsg


def main(argv):
    version = __version__
    configpath = ''
    statusmsg = {}
    debug=False
    jobtype = 'realtime'
    process = 'all'
    basepath = '/srv/archive'
    dayrange = 2
    destination='db'
    joblist = [] # jobtype can be realtime,second; archive;supergrad; etc
    newloggername = ''
    recentthreshold=7200
    sensorlist = []
    proxies = {}
    sendlog = False
    db = None
    telegramconfig = '/etc/martas/telegram.cfg'
    endtime = datetime.now(timezone.utc).replace(tzinfo=None)

    try:
        opts, args = getopt.getopt(argv,"hc:j:e:d:p:l:xD",["config=","joblist=","endtime=","dayrange=","path=","loggername","sendlog","debug=",])
    except getopt.GetoptError:
        print ('filter.py -c <config>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('-- filter.py to smooth and resample data --')
            print ('-----------------------------------------------------------------')
            print ('detailed description ..')
            print ('...')
            print ('...')
            print ('-------------------------------------')
            print ('Usage:')
            print ('python filter.py -c <config>')
            print ('-------------------------------------')
            print ('Options:')
            print ('-c (required) : configuration data path')
            print ('-j            : define jobtype i.e. realtime or archive')
            print ('-e            : endtime, default is now')
            print ('-d            : (int) dayrange, amount of days to analyze before endtime')
            print ('-p            : basepath - default is in config file')
            print ('-l            : loggername')
            print ('-------------------------------------')
            print ('Application:')
            print ('python3 filter.py -c ../conf/filter.cfg -j archive,second -d 3 -s LEMI025_28_0002_0001')
            sys.exit()
        elif opt in ("-c", "--config"):
            # delete any / at the end of the string
            configpath = os.path.abspath(arg)
        elif opt in ("-j", "--jobtype"):
            # define a jobtype (realtime,archive)
            joblist = arg.split(",")
            jobtype = joblist[0]
        elif opt in ("-d", "--dayrange"):
            # define a dayrange for archive jobs - default is 2
            dayrange = int(arg)
        elif opt in ("-e", "--endtime"):
            # define an endtime, default is now
            try:
                endtime = methods.testtime(arg)
            except:
                print (" Endtime could not be interpreted")
                endtime = datetime.now(timezone.utc).replace(tzinfo=None)
        elif opt in ("-p", "--path"):
            basepath = os.path.abspath(arg)
        elif opt in ("-l", "--loggername"):
            newloggername = arg
        elif opt in ("-x", "--sendlog"):
            sendlog = True
        elif opt in ("-D", "--debug"):
            print ("Debug activated ...")
            debug = True

    print ("Running filter.py version {} for {} job".format(version,jobtype))

    if not os.path.exists(configpath):
        print ('Specify a valid path to configuration information')
        print ('-- check filter.py -h for more options and requirements')
        sys.exit()

    if not jobtype == 'realtime':
        jobtype = "archive"

    if endtime:
        try:
            endtime = methods.testtime(endtime)
        except:
            print ("Endtime could not be interpreted - Aborting")
            sys.exit(1)


    print ("1. Read and check validity of configuration data")
    dd = read_conf(configpath)
    if not dd:
        print ("filter: empty configuration file - aborting")
        sys.exit(0)

    if debug:
        print ("Configuration:", dd)
    groupparameter = dd.get('groupparameter',{})
    permanent = dd.get('permanent',[])
    blacklist = dd.get('blacklist',[])
    basics = dd.get('basics',{})
    # By default the last 2 hours are filtered for realtime data
    try:
        recentthreshold=int(basics.get('recentthreshold',7200))
    except:
        recentthreshold=7200
    destination = basics.get('destination')
    basepath = basics.get('archivepath', "/srv/archive")
    receiver  = basics.get('notification',"")
    notificationcfg  = basics.get('notificationcfg',"")
    logpath = basics.get('logpath',"/tmp/filterstatus.log")
    if newloggername:
        logpath = os.path.join(os.path.dirname(logpath),newloggername)
    outputformat = basics.get('outputformat')
    credentials = basics.get('dbcredentials',"cobsdb")

    print ("2. Activate logging scheme as selected in config")
    #sn = 'ALDEBARAN' # servername  get name from machine ...
    sn = socket.gethostname()
    statusmsg = {}
    name = "{}-FILTER".format(sn)

    try:
        print ("Connecting to database")
        db = mm.connect_db(credentials)
        print ("... success")
        statusmsg[name] = 'DB on {} connected'.format(sn)
        # get existing Data IDs from data base
        sensorlist = db.select("DataID","DATAINFO")
    except:
        print ("... failed")
        statusmsg[name] = '{}: DB connection failed'.format(sn)

    statusmsg = apply_filter(db, statusmsg=statusmsg, groupdict=groupparameter, permanent=permanent, blacklist=blacklist, jobtype=jobtype, endtime=endtime, dayrange=dayrange, dbinputsensors=sensorlist, basepath=basepath, destination=destination, outputformat=outputformat, recentthreshold=recentthreshold, debug=debug)

    if basics.get('https'):
        proxies['https'] = basics.get('https')
    if basics.get('http'):
        proxies['http'] = basics.get('http')

    if not debug and sendlog:
        martaslog = ml(logfile=logpath,receiver=receiver,proxies=proxies)
        if receiver == 'telegram':
            martaslog.telegram['config'] = notificationcfg
        elif receiver == 'email':
            martaslog.email['config'] = notificationcfg
        martaslog.msg(statusmsg)
    elif debug:
        print ("Debug selected - statusmsg looks like:")
        print (statusmsg)

if __name__ == "__main__":
   main(sys.argv[1:])




"""
groupparameterdict   :   'LEMI' : {'station' : 'WIC', 'filtertype':'default', 'realtime' : True, 'window' : 40000},
                         'GSM90_14245_0002_0001' : {'filtertype':'gaussian', 'filterwidth':3.33333333, 'resample_period':1, 'realtime':True, 'window' : 10000},
                         'IWT_TILT01_0001_0003' : {'filtertype': 'default', 'realtime' : False, 'window' : 40000},
                         'BM35' : {'filtertype':'gaussian', 'filterwidth':3.33333333, 'resample_period':1, 'realtime':True, 'window' : 10000},
                         'G823A' : {'filtertype': 'default', 'realtime' : False, 'window' : 40000}}


# Realtime list is used for time slicing to avoid linearly filled gaps
realtime     :    LEMI025_22_0003_0001,LEMI036_1_0002_0001,LEMI036_3_0001_0001,GSM90_14245_0002_0001,BM35_029_0001_0001,BM35_033_0001_0001

blacklist    :    LEMI025_28_0002_0001,

scalardict   :    GP20S3NS_012201_0001_0001 : {'type' : 'GP20S3', 'components' : ['x','y','z']},
                  'GP20S3V_911005_0001_0001' :  {'type' : 'GP20S3', 'components' : ['x','y','z']},
                  'GP20S3EW_111201_0001_0001' :  {'type' : 'GP20S3', 'components' : ['x','y','z']}}
"""




# -*- coding: utf-8 -*-
"""
copyright:
    The MagPy Development Team
"""

# import order matters - NamedTemporaryFile must be one of the first!
__all__ = ['core','lib','app'
]

#import martas.core
import martas.app
#import magpy.core

#from .version import __version__



#!/usr/bin/env python
"""
a small tool for PalmAcq and ObsDAQ (Mingeo, Hungary)
palmobs.py has these options:
    forwarding mode
        get data from obsDAQ
    command mode
        talk to Palmdacq
    transparent mode
        talk to ObsDAQ
    -h for help
"""

from __future__ import print_function
import sys, time, os, socket, getopt
import serial
import struct, binascii, re, csv
from datetime import datetime, timedelta
from matplotlib.dates import date2num, num2date
import numpy as np
import time

from martas.core import methods as mm



# settings for PalmAcq
PORT = '/dev/ttyUSB0'
BAUDRATE='57600'
eol = '\r'
# since beginning of 2017 there are 18 leap seconds. PalmAcq starts with 15 leap seconds.
LEAPSECOND = 18

# settings for ObsDAQ
#   only baudrate 19200 supported here 
obsbaud = '19200'
escFromTranspChars = '\x00\x1b'
# GAINMAX is 10 for +/-10V (cc=02) and 5 for +/-5V (cc=03), see WS command
# needed only for option -o --output, no affect on data
GAINMAX = 10

# ----------------------------------
# please don't edit beyond this line
# ----------------------------------

global ser
global QUIET
QUIET = False

def lineread(ser,eol):
            # FUNCTION 'LINEREAD'
            # Does the same as readline() plus timeout 
            ser_str = ''
            timeout = time.time()+2
            if sys.version_info >= (3, 0):
                eol=eol.encode('ascii')
                while True:
                    byte = ser.read()
                    if byte == eol:
                        break
                    if time.time() > timeout:
                        print ('Timeout')
                        break
                    try:
                        ser_str += byte.decode('ascii')
                    except:
                        print('palmacq.py: decode error, got '+str(byte))
            else:
                while True:
                    char = ser.read()
                    if char == eol:
                        break
                    if time.time() > timeout:
                        print ('Timeout')
                        break
                    ser_str += char
            return ser_str

def send_command(ser,command,eol,hex=False):

    command = command+eol
    #print 'Command:  %s \n ' % command.replace(eol,'')
    sendtime = date2num(datetime.utcnow())
    #print "Sending"
    if sys.version_info >= (3, 0):
        ser.write(command.encode('ascii'))
    else:
        ser.write(command)
    #print "Received something - interpretation"
    response = lineread(ser,eol)
    #print "interprete"
    receivetime = date2num(datetime.utcnow())
    meantime = np.mean([receivetime,sendtime])
    #print "Timediff", (receivetime-sendtime)*3600*24
    return response, num2date(meantime).replace(tzinfo=None)


def command(call):
    global ser
    global QUIET
    if not QUIET:
        print(call)
    answer, actime = send_command(ser,call,eol)
    if not QUIET:
        print(answer)
    return answer



def main(argv):
    try:
        opts, args = getopt.getopt(argv,"hvm:tpfd:gsoiq",[])
    except getopt.GetoptError:
        print ('unknown option')
        sys.exit(2)

    global ser
    port = PORT
    baudrate = BAUDRATE
    targetleapsecond = LEAPSECOND
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('Sending data to PalmAcq resp. ObsDAQ')
            print ('-------------------------------------')
            print ('Usage:')
            print ('palmobs.py -q -v -t -m config-file -p -f -d [R/P/G] -g -s -i -o')
            print ('-------------------------------------')
            print ('Options:')
            print ('-v          : show version of PalmAcq and quit')
            print ('-m          : MARTAS compatible config file')
            print ('') 
            print ('-t          : enter Transparent mode')
            print ("-p          : exit Transparent mode, return to PalmAcq's Command mode")
            print ('-f          : enter Forwarding mode')
            print ('    -d R    : enter to ObsDAQ stream') 
            print ('    -d G    : enter to GPS info') 
            print ('    -d P    : enter to PPS') 
            print ('-g          : get GPS leap seconds')
            print ('') 
            print ('-s          : show output from serial line')
            print ('-o          : show the formatted output')
            print ('-i          : show info about PalmAcq settings')
            print ("-q          : quiet: don't show commands and answers. Has to be first option.")

            print ('-------------------------------------')
            print ('Examples:')
            print ('python palmobs.py -f R')
            sys.exit()
        if opt in ("-m"):
            configfile = os.path.abspath(arg)
            conf = mm.get_conf(configfile)
            port = conf.get('port')
            if 'LEAPSECOND' in conf:
                try:
                    targetleapsecond = int(conf.get('LEAPSECOND'))
                except:
                    print('could not read LEAPSECOND from config file '+configfile)

        ser = serial.Serial(port, baudrate=baudrate , parity='N', bytesize=8, stopbits=1, timeout=2)

        if opt in ("-q", "--quiet"):
            global QUIET
            QUIET = True
        if opt in ("-v", "--version"):
            # e.g. PALMACQ fw. v4.2.2 SM
            command('GV')
            # there come two lines, this cannot be handled here correctly
            print (lineread(ser,eol))
            # programmed not perfectly, so better quit here...
            #quit() 
        elif opt in ("-t", "--transparent"):
            # Transparent mode - this connects only to RS-485 port
            command('SP:R:'+obsbaud+',8,n,1,1')
            command('ST:R')
            a=command('SM:TRP')
            if a=='AM:TRP,001B':
                print ("PalmAcq: going to transparent mode...")
            else:
                # this does not mean, that the connection wasn't made
                # due to a timing problem
                print ("PalmAcq: probably no connection when trying to enter transparent mode")
            # print firmware version and firmware date
            command('$01F')
        elif opt in ("-p", "--program"):
            print ('PalmAcq: trying to enter command mode')
            # return from Transparent mode to PalmDAQ's Command mode
            if sys.version_info >= (3, 0):
                ser.write(escFromTranspChars.encode('ascii'))
            else:
                ser.write(escFromTranspChars)
            command('SM:CMD')
            # if in Forward mode, return from there, if not, doesn't matter
            command('SF:-')
        elif opt in ("-f", "--forward"):
            command('SF')
        elif opt in ("-d"):
            if arg == '':
                command('SF')
            elif arg in ['R','P','G','M','C']:
                command('SF:'+arg)
                print ('PalmAcq: entering forward mode "'+arg+'"')
            else:
                print ('PalmAcq: forward mode: SF: '+arg+': bad argument')
                quit()
        elif opt in ("-g", "--gpsinfo"):
            # GPS sends amount of leap seconds, 2017 it was 18
            leapsecond = 0
            while leapsecond < targetleapsecond:
                command('TP:G:L,$PMTK457*34')
                command('GB')
                result = ''
                while not result == ':END':
                    result = lineread(ser,eol)
                    l = result.split('$PMTK557,')
                    if len(l) == 2:
                        leapsecond = int(l[1].split('.')[0])
                        print ('PalmAcq: Leapsecond according to GPS signal: '+ str(leapsecond))
                time.sleep(2)
                if leapsecond < targetleapsecond:
                    print ('..trying once again')
        elif opt in ("-i", "--info"):
            command('GI:PORTS')
            command('GI:MC')
            command('GI:TF')
            command('GI:MLL')
            command('GI:BUFF')
            command('GI:SCH')
            command('GI:BC')
            command('GI:CRD')
            command('GI:GPS')
            quit()
        elif opt in ("-s", "--show"):
            while True:
                print (lineread(ser,eol))
            sys.exit()
        elif opt in ("-o", "--output"):
            while True:
                l = lineread(ser,eol).strip()
                if l.startswith(':R'):
                    # :R,00,200131.143739.617,*0259FEFFF1BFFFFCEDL:04AC11CC000B000B000B
                    # :R,00,YYMMDD.hhmmss.sss,*xxxxxxyyyyyyzzzzzzt:vvvvttttppppqqqqrrrr
                    d = l.split(',')
                    Y = int('20'+d[2][0:2])
                    M = int(d[2][2:4])
                    D = int(d[2][4:6])
                    h = int(d[2][7:9])
                    m = int(d[2][9:11])
                    s = int(d[2][11:13])
                    us = int(d[2][14:17]) * 1000
                    timestamp = datetime(Y,M,D,h,m,s,us)
                    if d[3][0] == '*':
                        x = (int('0x'+d[3][1:7],16) ^ 0x800000) - 0x800000
                        x = float(x) * 2**-23 * GAINMAX
                        y = (int('0x'+d[3][7:13],16) ^ 0x800000) - 0x800000
                        y = float(y) * 2**-23 * GAINMAX
                        z = (int('0x'+d[3][13:19],16) ^ 0x800000) - 0x800000
                        z = float(z) * 2**-23 * GAINMAX
                        triggerflag = d[3][19]
                    else:
                        pass
                    sup = d[3].split(':')
                    if len(sup) == 2:
                        voltage = int(sup[1][0:4],16) ^ 0x8000 - 0x8000
                        voltage = float(voltage) * 2.6622e-3 + 9.15
                        temp = int(sup[1][4:8],16) ^ 0x8000 - 0x8000
                        temp = float(temp) / 128.
                        p = (int('0x'+sup[1][8:12],16) ^ 0x8000) - 0x8000
                        p = float(p) / 8000.0
                        q = (int('0x'+sup[1][12:16],16) ^ 0x8000) - 0x8000
                        q = float(q) / 8000.0
                        r = (int('0x'+sup[1][16:20],16) ^ 0x8000) - 0x8000
                        r = float(r) / 8000.0
                    print (str(timestamp)+'\t',end='')
                    print (str(x)+'\t',end='')
                    print (str(y)+'\t',end='')
                    print (str(z)+'\t',end='')
                    print (str(triggerflag))
                    if len(sup) == 2:
                        print ('supplementary:\t',end='')
                        print (str(voltage)+' V\t',end='')
                        print (str(temp)+' degC\t',end='')
                        print (str(p)+'\t',end='')
                        print (str(q)+'\t',end='')
                        print (str(r)+'\t')
                else:
                    print (l)
            sys.exit()

if __name__ == "__main__":
    main(sys.argv[1:])





# ###################################################################
# Import packages
# ###################################################################

import time
# ###################################################################
# Import packages
# ###################################################################

import serial
import time
import sys
import operator
from functools import reduce
from datetime import datetime, timezone

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection

from magpy.core.methods import is_number
from twisted.python import log
from martas.core import methods as mm
import subprocess

## GNSSHAT protocol
## -----------------------

class GNSSHATProtocol(object):
    """
    Protocol to read Waveshare GNSS HAT data attached to a raspberry.
    https://www.waveshare.com/wiki/GSM/GPRS/GNSS_HAT

    Please note: you might need to reset the hat on the side button next to the gnss antenna

    """

    def __init__(self, client, sensordict, confdict):
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.counter = {}
        self.ser = None

        # used are currently on command 0 and 4
        self.commands = ["AT+CGNSPWR=1", "AT+CGNSSEQ=\"RMC\"", "AT+CGNSINF", "AT+CGNSURC=2", "AT+CGNSTST=1"]
        self.eol = "\r\n"

        # Serial communication data (from confdict)
        self.baudrate=int(sensordict.get('baudrate'))
        self.port = confdict['serialport']+sensordict.get('port')
        self.parity=sensordict.get('parity')
        self.bytesize=sensordict.get('bytesize')
        self.stopbits=sensordict.get('stopbits')
        self.timeout=2 # should be rate depended

        # extract data from configuration dictionary
        debugtest = confdict.get('debug')

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # switch on debug mode
        self.debug = False
        if debugtest in ['true','True','TRUE']:
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))


    def datetime2array(self,t):
        return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

    def serwrite(self, ser, command, eol="\r\n"):
        fc = command + eol
        ffc = fc.encode()
        if sys.version.startswith('3'):
            ser.write(ffc)
        else:
            ser.write(fc)

    def open_connection(self):
        log.msg("connecting to serial port")
        ser = serial.Serial(self.port, baudrate=int(self.baudrate), parity=self.parity, bytesize=int(self.bytesize), stopbits=int(self.stopbits), timeout=self.timeout)
        time.sleep(2)
        self.serwrite(ser, self.commands[0])
        ser.flushInput()
        self.data = b""
        self.num = 0
        self.ser = ser
        return ser


    def processGNSSData(self, sensorid, res):
        """
        res =
        {'time': '134353.000', 'lat': 48.08448035, 'lon': 10.848522516666666, 'date': '150226', 'status': 'A',
        'speed': '1.29', 'course': '110.32', 'declination': '', 'dec_orient': '', 'integrity': 'A*72',
        'datetime': '2026-02-15T13:43:54.000000', 'quality': '1', 'N': '5', 'horizontal_prescision': '1.23',
        'altitude': '583.395', 'altitude_unit': 'M', 'geoidal_separation': '47.864',
        'geoidal_separation_unit': 'M', 'age_dgps': '', 'dgps_reference': '*4A'}
        """
        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        gnsstime = res.get("datetime", currenttime)

        if self.sensordict.get('ptime','') in ['NTP','ntp']:
            secondtime = gnsstime
            maintime = currenttime
            sectimename = "GPS"
        else:
            maintime = gnsstime
            secondtime = currenttime
            sectimename = "NTP"

        outdate = datetime.strftime(maintime,"%Y-%m-%d")
        filename = outdate

        datearray = self.datetime2array(maintime)

        fullkeylist = ['lat', 'lon']  # 1000000000
        numkeylist = ['altitude', 'speed', 'course', 'horizontal_prescision', 'geoidal_separation'] # 1000
        intkeylist = ['quality', 'N', 'status'] # 1
        #strkeylist = ['dgps_reference', 'status', 'integrity']
        keys = '[x,y,z,f,var1,dx,dz,var2,var3,var4,sectime]'

        packcode = '6hL'

        values = []
        multiplier = []
        unit = []
        data_bin = None

        for key in fullkeylist:
            dat = res.get(key, 0.000)
            try:
                values.append(float(dat))
                datearray.append(int(float(dat)*1000000000))
                packcode = packcode + 'q'
                multiplier.append(1000000000)
                unit.append("deg")
            except:
                log.msg('{} protocol: Error while appending data to file (non-float?): {}'.format(self.sensordict.get('protocol'),dat) )
        for key in numkeylist:
            dat = res.get(key, 0.000)
            if not is_number(dat):
                dat = 0.000
            try:
                values.append(float(dat))
                datearray.append(int(float(dat)*1000))
                packcode = packcode + 'l'
                multiplier.append(1000)
                if key == "altitude":
                    unit.append(res.get('altitude_unit','').lower())
                elif key == "geoidal_separation":
                    unit.append(res.get('geoidal_separation_unit', '').lower())
                else:
                    unit.append("")
            except:
                log.msg('{} protocol: Error while appending data to file (non-float?): {}'.format(self.sensordict.get('protocol'),dat) )
        for key in intkeylist:
            dat = res.get(key, 0)
            try:
                if key == 'status':
                    if dat == 'A':
                        dat = 1
                    else:
                        dat = 0
                values.append(int(dat))
                datearray.append(int(dat))
                packcode = packcode + 'l'
                multiplier.append(1)
                unit.append("")
            except:
                log.msg('{} protocol: Error while appending data to file (non-float?): {}'.format(self.sensordict.get('protocol'),dat) )

        packcode = packcode + '6hL'
        datearray.extend(self.datetime2array(secondtime))

        try:
            data_bin = struct.pack('<'+packcode,*datearray) #little endian
        except:
            log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))
            pass

        intkeylist.append('{}time'.format(sectimename.lower()))
        unit.append("")
        multiplier.append(0)
        ele = str(fullkeylist+numkeylist+intkeylist).replace(" ","").replace("'", "").strip()
        unit = str(unit).replace(" ","").replace("'", "").strip()
        multplier = str(multiplier).replace(" ","")
        # Correct some common old problem
        unit = unit.replace('deg C', 'degC')
        #print (ele,multiplier,unit)

        header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, keys, ele, unit, multplier, packcode, struct.calcsize(packcode))

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)

        return ','.join(list(map(str,datearray))), header

    def get_nmea(self, datastr, debug=False):
        """
        Extract dictionaries from RMC and GGA Strings
        RMC:
        GGA:
        When returning: GGA inputs will overwrite similar key from RMC
        """
        lines = datastr.split()
        rd, gd = {}, {}
        for line in lines:
            if line.startswith('$GNGGA'):
                if debug:
                    print("GGA:", line)
                gd = self.analyse_gga(line)
            if line.startswith('$GNRMC'):
                if debug:
                    print("RMC:", line)
                rd = self.analyse_rmc(line)
        return {**rd, **gd}

    def checksum_nmea(self, line):
        """
        Checking NMEA strings by checksum
        """
        tline = line.strip("$\n")
        try:
            nmea, cont_checksum = tline.split("*", 1)
            calc_checksum = reduce(operator.xor, (ord(s) for s in nmea), 0)
            if int(cont_checksum, base=16) == calc_checksum:
                return True
            else:
                print("NMEA: calculated and contained checksums differ")
                return False
        except:
            print("NMEA: checksums problem")
            return False

    def convert_ddmm(self, ddmm, orientation="", debug=False):
        """
        latitude/loingitude format: (D)DDMM.MMMMM
        orientation = N/S or E/W
        """
        sign = 1

        if not ddmm in ['']:
            DD = int(float(ddmm) / 100)
            MM = float(ddmm) - DD * 100
            if debug:
                print(DD)
                print(MM)
            if orientation in ["S", "W"]:
                sign = -1
            decimals = sign * (DD + MM / 60.)
        else:
            decimals = 999.99

        return decimals

    def analyse_gga(self, ggaline):
        ggasplit = ggaline.split(',')
        gd = {}
        if len(ggasplit) >= 15 and self.checksum_nmea(ggaline):
            gd['time'] = ggasplit[1]
            gd['lat'] = self.convert_ddmm(ggasplit[2], ggasplit[3])
            gd['lon'] = self.convert_ddmm(ggasplit[4], ggasplit[5])
            gd['quality'] = ggasplit[6]
            gd['N'] = ggasplit[7]
            gd['horizontal_prescision'] = ggasplit[8]
            gd['altitude'] = ggasplit[9]
            gd['altitude_unit'] = ggasplit[10]
            gd['geoidal_separation'] = ggasplit[11]
            gd['geoidal_separation_unit'] = ggasplit[12]
            gd['age_dgps'] = ggasplit[13]
            gd['dgps_reference'] = ggasplit[14]
            # gd['checksum'] = ggasplit[15]
        return gd

    def analyse_rmc(self, rmcline):
        rmcsplit = rmcline.split(',')
        rd = {}
        if len(rmcsplit) >= 13 and self.checksum_nmea(rmcline):
            rd['time'] = rmcsplit[1]
            rd['lat'] = self.convert_ddmm(rmcsplit[3], rmcsplit[4])
            rd['lon'] = self.convert_ddmm(rmcsplit[5], rmcsplit[6])
            rd['date'] = rmcsplit[9]
            rd['status'] = rmcsplit[2]
            rd['speed'] = rmcsplit[7]
            rd['course'] = rmcsplit[8]
            rd['declination'] = rmcsplit[10]
            rd['dec_orient'] = rmcsplit[11]
            rd['integrity'] = rmcsplit[12]
            datetime_str = "{} {}".format(rmcsplit[9], rmcsplit[1])
            datetime_obj = datetime.strptime(datetime_str, '%d%m%y %H%M%S.%f')
            #rd['datetime'] = datetime.strftime(datetime_obj, '%Y-%m-%dT%H:%M:%S.%f')
            rd['datetime'] = datetime_obj
        return rd


    def sendRequest(self):
        # connect to serial
        ser = self.ser
        if not ser or not ser.isOpen():
            ser = self.open_connection()

        while ser.inWaiting() > 0:
            self.data += ser.read(ser.inWaiting())
        #print ("data", data.decode())
        if not self.data == b"":
            try:
                result = self.data.decode()
                res = self.get_nmea(result)
            except:
                res = {}
            if self.debug:
                print ("Obtained:", res)
            time.sleep(0.5)
            self.num = self.num + 1
            if self.num > 4:
                self.num = 4
            self.serwrite(ser,self.commands[self.num])
            self.data = b""
            #publish it
            self.publish_data(res)

    def publish_data(self, res):

        if res and len(res) > 0:
            topic = self.confdict.get('station') + '/' + self.sensor
            pdata, head = self.processGNSSData(self.sensor, res)

            senddata = False
            try:
                coll = int(self.sensordict.get('stack'))
            except:
                coll = 0

            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(pdata)
                    self.datacnt += 1
                else:
                    senddata = True
                    pdata = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", pdata, qos=self.qos)
                # If multiple sensors are connected, self.count needs to be a dict (with topic)
                # Initialize counter for each topic
                if self.counter.get(self.sensor,'') == '':
                    self.counter[self.sensor] = 0
                cnt = self.counter.get(self.sensor)

                if cnt == 0:
                    ## 'Add' is a string containing dict info like:
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)

                cnt += 1
                #self.count += 1
                if cnt >= self.metacnt:
                    cnt = 0

                # update counter in dict
                self.counter[self.sensor] = cnt



#!/usr/bin/env python

from magpy.stream import *
from martas.version import __version__
import martas.core.methods as mm
import dash
from dash import html, dash_table, dcc, Output, Input, callback
import dash_daq as daq
from plotly.subplots import make_subplots
from martas.app.monitor import _latestfile
import random
from crontab import CronTab
import psutil
from pathlib import Path
import numpy as np

"""
DESCRIPTION:
a visual summary of a MARCOS server contains:
- table of sensorIDs plus datatables and actuality
  i.e. LEMI025_22_0004   |     | _0001 (green), _0002(green) |  location: StationID, PierID
- collect processes (list of collector processes) and health state
  i.e. PROMETHEUS |  status  | sensorA (green), sensorB (green) | location: StationID, PierID 
- collect external data source data actuality (list of collector processes) and health state
  i.e. GOES17 |  green |  dataset  | source location 
- a general health state section
  i.e. MARCOS on, essential jobs on green, secondary jobs on green, SUCCESS in all scheduled jobs
  accessibility of Hohe Warte
- status of scheduler and time to next job running
- Status Lamps for important thresholds
- local storage space (gauge)
- selection of three timeseries live plots with component selection
- group, type, element selector with current values (LED) and activity status


Required packages
dash
plotly
pip install dash_daq
"""
# Configuration information
mainpath = os.path.dirname(os.path.realpath(__file__))
configpath = os.path.join(mainpath, "", "..", "..", "conf", "archive.cfg")
webconfigpath = os.path.join(mainpath, "..", "..", "conf", "web.cfg")

# Read configuration data and initialize amount of plots
cfg = mm.get_conf(configpath)
webcfg = mm.get_conf(webconfigpath)
archivepath = cfg.get('archivepath')
logpath = cfg.get('logpath')
dbcred = cfg.get('dbcredentials')

tester = True
if tester:
    print(cfg)
    archivepath = "/srv"
    logpath = "/home/leon/.martas/log"
    dbcred = 'cobsdb'
    configpath = "/home/leon/.martas/conf/archive.cfg"

# import a webpage config: darkmode, defaultpage, default limits, displayrate, read file first
defaultpage = None  # set default # get from config
debug = False
if webcfg.get('defaultpage','martas') in ['marcos','Marcos','MARCOS']:
    defaultpage = "/" # set default # get from config
srate = int(webcfg.get('refreshrate',5)) # displayrate - needs to be large enough, dynamically adjusted
if webcfg.get('read_initial_buffer',False) in ['False','false','FALSE', False]:
    read_initial_buffer = False
if webcfg.get('debug',False) in ['True','true','TRUE', True]:
    debug = True


statusdict = {"archive" : {"space" : 400, "used": 150, "cronenabled": False, "active": False, "logstatus":False },
              "monitor" : {"space" : 100, "used": 100, "cronenabled": False, "active": False, "logstatus":False },
              "database" : {"space" : 300, "used": 250, "cronenabled": False, "active": False, "logstatus":False }
              }


def get_datainfo_from_db(cred='cobsdb', debug=False):
    """
    DESCRIPTION
        select all datatables from database, check DATAINFO, check contents
    RETURNS
        dictionary with DataID, SensorID, datatable ok, DATAINFO ok, last input, first input, StationID, PierID
    """

    db = mm.connect_db(cred, False, False)
    if db:
        dbd = statusdict.get('database')
        dbd['active'] = True
        statusdict['database'] = dbd

    def _analyse_columns(colsstr,unitsstr):
        usedkeys, components, counits = [],[],[]
        cols, units = [], []
        if colsstr:
            cols = colsstr.split(",")
        if unitsstr:
            units = unitsstr.split(",")
        allkeys = DataStream().KEYLIST
        for i,co in enumerate(cols):
            if co:
                components.append(co)
                if len(units)>=i:
                   counits.append(units[i])
                usedkeys.append(allkeys[i+1])
        return usedkeys, components, counits

    def _get_column_names(tablename):
        keys = []
        sql = "SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '{}'".format(tablename)
        cursor = db.db.cursor()
        msg = db._executesql(cursor, sql)
        n = cursor.fetchall()
        for l in n:
            keys.append(l[3])
        # drop time column and only return keys with more then time
        if len(keys) > 1:
            return keys[1:]
        else:
            return []

    # Check whether DB still available
    result = {}
    missingdatainfo = []
    now = datetime.now(timezone.utc).replace(tzinfo=None)

    if db:
        output = db.select('DataID,SensorID,StationID,ColumnContents,ColumnUnits,DataPier', 'DATAINFO')
        for elem in output:
            #print ("Columns", elem[3],elem[4])
            usedkeys, components, units = _analyse_columns(elem[3],elem[4])
            result[elem[0]] = {"SensorID":elem[1],"StationID":elem[2],"PierID":elem[5],"DataKeys":usedkeys,"DataElements":components,"DataUnits":units,"TableExists":False,"DataInfoExists":True,"FirstInput":None,"LastInput":None,"Actual":0}
        sql = "SHOW TABLES"
        cursor = db.db.cursor()
        msg = db._executesql(cursor, sql)
        n = cursor.fetchall()
        for elem in n:
            tablename = elem[0]
            if tablename.count("_") == 3:
                #follows the naming convention of MagPy
                cont = result.get(tablename, None)
                if cont:
                    cont["TableExists"] = True
                    result[tablename] = cont
                else:
                    missingdatainfo.append(tablename)
        for elem in missingdatainfo:
            # Get columns of existing data
            keys = _get_column_names(elem)
            if len(keys) > 0:
                result[elem] = {"SensorID":elem[:-5],"StationID":"","PierID":"","DataKeys":keys,"TableExists":True,"DataInfoExists":False,"FirstInput":None,"LastInput":None,"Actual":0}
            else:
                if debug:
                    print("ERROR with table {}: no data keys".format(elem))

        # Update start and enddate for all DataID
        for elem in result:
            cont = result.get(elem)
            #print (elem)
            if cont.get("TableExists"):
                try:
                    l = db.get_lines(elem, 1)
                    cont['LastInput'] = l.end()
                    diff = (now - l.end()).total_seconds()
                    if diff < 3600: # Threshold corresponds to maximal graph range - only those can be plotted
                        cont["Actual"] = 2
                    elif diff < 87000: # data younger then 1 day
                        cont["Actual"] = 1
                    cont['TimeDiff'] = diff
                    result[elem] = cont
                except:
                    if debug:
                        print ("ERROR with get_lines in table {}".format(elem))
                    pass

        if debug:
            print ("Summary", result)
        return result


def convert_datainfo_to_datatable(result, debug=False):
    """
    DESCRIPTION
        converts results dictionary into a Table for display
    """
    if not result:
        result = {}
    table = []
    dtable = []
    sidtab = []
    for elem in result:
        cont = result.get(elem)
        sid = cont.get("SensorID")
        did = elem[-5:]
        if sid in sidtab:
            line = [s for s in table if s[0] == sid][0]
            if debug:
                print ("found existing line", line)
            if cont.get("Actual") > line[1]:
                line[1] = cont.get("Actual")
            line[2].append(did)
        else:
            table.append([sid, cont.get("Actual"), [did], cont.get("StationID"), cont.get("PierID")])
        sidtab.append(sid)

    for line in table:
        dtable.append({"SensorID" : line[0], "Actual" : line[1], "DataIDs" : ",".join(line[2]), "StationID" : line[3], "PierID" : line[4]})
    dcols = ["SensorID", "Actual", "DataIDs", "StationID", "PierID"]
    return dtable, dcols


def convert_datainfo_to_idtable(result, debug=False):
    """
    DESCRIPTION
        converts results dictionary into a Table for display
    """
    dtable = []
    for elem in result:
        cont = result.get(elem)
        comps = cont.get("DataElements",[])
        if len(comps) > 8:
            comps = comps[:8]
            comps.append("...")
        comps = ",".join(comps)
        dtable.append({"DataID" : elem, "in DATAINFO" : str(cont.get("DataInfoExists")), "Data table" : str(cont.get("TableExists")), "Actual" : cont.get("Actual"), "Components" : comps})
    dcols = ["DataID", "in DATAINFO", "Data table", "Actual", "Components"]
    return dtable, dcols


def get_graph_options(result, actualcrit=1):
    """
    DESCRIPTION
        get a list of actual sensors for dropdown and available keys/elements for plotting
    """
    if not result:
        result = {}
    dataoptions = []
    datavalue = None
    for elem in result:
        cont = result.get(elem)
        actual = cont.get("Actual")
        if actual >= actualcrit:
            dataoptions.append(elem)
    if len(dataoptions) > 0:
        datavalue = random.choice(dataoptions)
    return dataoptions, datavalue


def get_graph_keys(tablename, result):
    """
    DESCRIPTION
        get a list of keys for dropdown selection
    """
    keyoptions = [] # for datavalue
    keyvalue = None # for datavalue
    if not tablename:
        return [],""

    cont = result.get(tablename)
    opt = []
    keys = cont.get("DataKeys",[])
    choice = random.choice(keys)
    units = cont.get("DataUnits",[])
    comps = cont.get("DataElements",[])
    for i,key in enumerate(keys):
        if len(comps) > 0 and len(comps) >= i and len(units) >= i and units[i]:
            label = "{} [{}]".format(comps[i], units[i])
        elif len(comps) > 0 and len(comps) >= i:
            label = "{}".format(comps[i])
        else:
            label = "{}".format(key)
        opt.append({'label': label, "value":key })
    keyoptions = opt
    keyvalue = choice

    return keyoptions, keyvalue


def get_data(datatable, keys, datainfo=None, duration=60, cred="cobsdb"):
    mydata = {}
    names = []
    db = mm.connect_db(cred, False, False)
    # This job needs including trim needs about 1 sec on my comp
    stream = db.get_lines(datatable, 36000)
    endtime = stream.timerange()[1]
    stream = stream.trim(starttime=endtime-timedelta(minutes=duration), endtime=endtime)
    mydata['time'] = stream.ndarray[0]
    for key in keys:
        name = key
        #name = stream.get_key_name(key) # get name from results
        mydata[name] = stream._get_column(key)
        names.append(name)
    return mydata, names


# slow methods
def getspace(path="/srv"): # path = '/srv'
    """
    DESCRIPTION
        get space from monitor
    """
    statvfs = os.statvfs(path)
    total = (statvfs.f_frsize * statvfs.f_blocks / (1024.*1024.))     # Size of filesystem in bytes
    remain = (statvfs.f_frsize * statvfs.f_bavail / (1024.*1024.))     # Number of free bytes that ordinary users
    usedper=100-(remain/total*100.)
    used = total-remain
    return total, used


def get_storage_usage(statusdict=None, cred="cobsdb", archivepath="", logpath="", debug=False):
    if not statusdict:
        statusdict = {}
    txt = ""
    dbused = 0
    db = mm.connect_db(cred, False, False)
    if db:
        # This line creates a stdout, might be changed to only a return value
        txt = db.info(destination='stdout', level='full')
    if txt:
        dbused = txt.split()[-1]
    dbmax = 30.0 # should be smaller than 30GB
    atotal, aused = 0, 0
    mtotal, mused = 0, 0
    if archivepath:
        if os.path.exists(archivepath):
            atotal, aused = getspace(path=archivepath)
    if logpath:
        logpath = os.path.dirname(logpath)
        if os.path.exists(logpath):
            mtotal, mused = getspace(path=logpath)
    ad = statusdict.get('archive',{})
    if ad:
        ad["space"] = np.round(atotal/1000.,0)
        ad["used"] = np.round(aused/1000.,0)
    md = statusdict.get('monitor',{})
    if md:
        md["space"] = np.round(mtotal/1000.,0)
        md["used"] = np.round(mused/1000.,0)
    if debug:
        print ("SPACE", atotal, aused)
        print("SPACE", mtotal, mused)
    dd = statusdict.get('database',{})
    if dd:
        dd["space"] = dbmax
        dd["used"] = float(dbused)
    return statusdict


def get_pid(name):
    pid = 0
    for proc in psutil.process_iter(attrs=["pid", "name", "exe", "cmdline"]):
        if isinstance(proc.info.get('cmdline'), (list,tuple)):
            for cmd in proc.info.get('cmdline'):
                if name in cmd:
                    pid = proc.pid
                    break
            if pid:
                break
    return pid


def get_cron_jobs(statusdict=None,cred="cobsdb", archivepath="", logpath="",debug=False):
    """
    DESCRIPTION
        get active Cron jobs once per minute
    """
    if not statusdict:
        statusdict = {}

    statusdict = get_storage_usage(statusdict, cred=cred, archivepath=archivepath, logpath=logpath, debug=False)
    crontable = []
    marcoslist = []

    mycron = CronTab(user=True)
    for job in mycron:
        if debug:
            print ("Testing job:", job)
        comm = job.comment
        cand = job.command
        en = job.is_enabled()
        cl = cand.split()
        if comm.find("Running MARCOS") >= 0:
            logstatus = False
            active = False
            k = comm.replace("Running MARCOS process ","")
            pidname = Path(cl[2]).stem
            if debug:
                print ("Testing pid", pidname)
            p = get_pid(pidname)
            if debug:
                print (" - got pid", p)
            if p > 0:
                active = True
            for el in cl:
                if el.find(".log") >= 0:
                    te = el.split("/")
                    if debug:
                        print ("Found MARCOS job", te)
                    try:
                        ctime = os.path.getctime(el)
                        ctime = datetime.fromtimestamp(ctime)
                        logstatus = True
                    except:
                        pass
            statusdict[k] = {"cronenabled": en, "active": active, "logstatus": logstatus}
            marcoslist.append(k)
        else:
            logstatus = False
            active = False
            jobcommand = cl[1]
            lf = ""
            ctime = ""
            if len(jobcommand) < 3:
                jobcommand = cl[2]
            jc = jobcommand.split("/")[-1]
            for el in cl:
                if el.find(".log") >= 0:
                    te = el.split("/")
                    if debug:
                        print ("Found basic cron job", te)
                    try:
                        ctime = os.path.getctime(el)
                        ctime = datetime.fromtimestamp(ctime)
                        logstatus = True
                        if ctime > datetime.now()-timedelta(days=8):
                            active = True
                        ctime = ctime.strftime("%Y-%m-%dT%H:%M:%S")
                    except:
                        logstatus = False
                        active = False
                    try:
                        jn = jc.split(".")[0]
                        if jn in ["archive", "monitor"]:
                            jnstat = statusdict.get(jn,{})
                            jnstat["cronenabled"] = en
                            jnstat["active"] = active
                            jnstat["logstatus"] = logstatus
                            statusdict[jn] = jnstat
                        elif jn in ["db_truncate"]:
                            dbstat = statusdict.get("database",{})
                            dbstat["cronenabled"] = en
                            dbstat["logstatus"] = logstatus
                            statusdict["database"] = dbstat
                    except:
                        pass
                    lf = te[-1]
            resd = {"job" : jc, "enabled":str(en), "logfile":lf, "last log":ctime}
            crontable.append(resd)
    croncols = ["job", "enabled", "logfile", "last log"]

    return crontable, croncols, marcoslist, statusdict


def get_marcos_html(statusdict,marcos,i):
    htmllist = []
    oncolor = '#008003'
    offcolor = '#FF5E5E'
    acolor = offcolor
    ecolor = offcolor
    lcolor = offcolor
    if statusdict[marcos].get('active'):
        acolor = oncolor
    if statusdict[marcos].get('cronenabled'):
        ecolor = oncolor
    if statusdict[marcos].get('logstatus'):
        lcolor = oncolor
    #html.Ul([html.Li("Item 1"), html.Li("Item 2")])

    htmllist.append(html.Tr([html.Td(
        daq.Indicator(id='marcos{}active'.format(i), value=True, color=acolor,
                      style={"float": "left", "margin-right": "10px"})),
        html.Td(html.B("{}".format(marcos)))]
    ))
    htmllist.append(html.Tr([html.Td(
        daq.Indicator(id='marcos{}enabled'.format(i), value=True, color=ecolor,
                      style={"float": "left", "margin-right": "10px"})),
        html.Td(html.P("scheduled"))],
    ))
    htmllist.append(html.Tr([html.Td(
        daq.Indicator(id='marcos{}log'.format(i), value=True, color=lcolor,
                      style={"float": "left", "margin-right": "10px"})),
        html.Td(html.P("log status"))],
    ))
    return html.Table(htmllist)


# Initialize basic result dictionary (fast interval, graph and table)
result = get_datainfo_from_db(cred=dbcred, debug=debug)
dtable, dcols = convert_datainfo_to_datatable(result)
dataoptions, datavalue = get_graph_options(result)
keyoptions, keyvalue = get_graph_keys(datavalue, result)

# Initialize status dictionary (slow interval, diskspace, cron and processes)
crontable, croncols, marcoslist, statusdict = get_cron_jobs(statusdict=statusdict, cred=dbcred, archivepath=archivepath, logpath=logpath, debug=debug)

dash.register_page(__name__, path=defaultpage, top_nav=True, assets_ignore='martas.css')
#app = Dash(__name__, assets_ignore='martas.css')

layout = (html.Div(
                 className="mwrap",
                 children=[
                              html.Div(
                                  className="mbox ma",
                                  children=html.Div([
                                      html.Img(src='/assets/header.png',style={'width': '60%'}),
                                      ])
                              ),
                              html.Div(
                                  className="mbox mb",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos1')
                                      ])
                              ),
                              html.Div(
                                  className="mbox mc",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos2')
                                      ])
                              ),
                              html.Div(
                                  className="mbox md",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos3')
                                      ])
                              ),
                              html.Div(
                                  className="mbox me",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos4')
                                      ])
                              ),
                              html.Div(
                                  className="mbox mf",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos5')
                                      ])
                              ),
                              html.Div(
                                  className="mbox mg",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos6')
                                      ])
                              ),
                              html.Div(
                                  className="mbox mh",
                                  children=html.Div([
                                      html.Div(id='live-update-marcos7')
                                      ])
                              ),
                             html.Div(
                                 className="mbox mi",
                                 children=html.Div([
                                     html.P('Time coverage (minutes):'),
                                     dcc.Slider(id='durationslider', min=5, max=60, step=5, value=10,
                                                marks={x: str(x) for x in [10, 20, 30, 40, 50, 60]}),
                                     html.P('Height of the graph display (px):'),
                                     dcc.Slider(id='heightslider', min=200, max=1000, step=25, value=450,
                                                marks={x: str(x) for x in [200, 400, 600, 800, 1000]}),
                                 ])
                             ),
                             html.Div(
                                  className="mbox mj",
                                  children=html.Div([
                                      html.Div(id='live-update-dbstatus')
                                      ])
                              ),
                              html.Div(
                                  className="mbox mk",
                                  children=html.Div([
                                      html.Div(id='live-update-monitorstatus')
                                      ])
                              ),
                              html.Div(
                                  className="mbox ml",
                                  children=html.Div([
                                      html.Div(id='live-update-archivestatus')
                                      ])
                              ),
                              html.Div(
                                  className="mbox mm",
                                  children=html.Div([
                                      daq.Gauge(
                                          id='db-gauge',
                                          units="GB",
                                          size=120,
                                          color={"gradient": True,
                                                 "ranges": {"green": [0, statusdict['database'].get('space')*0.6], "yellow": [statusdict['database'].get('space')*0.6, statusdict['database'].get('space')*0.8], "red": [statusdict['database'].get('space')*0.8, statusdict['database'].get('space')]}},
                                          value=statusdict['database'].get('used'),
                                          label='DB storage',
                                          max=statusdict['database'].get('space'),
                                          min=0,
                                      ),
                                      dcc.Interval(
                                          id='gauge-update',
                                          interval=60 * 1000,  # in milliseconds
                                          n_intervals=0
                                      )
                                  ])
                              ),
                              html.Div(
                                  className="mbox mn",
                                  children=html.Div([
                                    daq.Gauge(
                                        id='base-gauge',
                                        size=120,
                                        color={"gradient": True,
                                               "ranges": {"green": [0, statusdict['monitor'].get('space') * 0.8],
                                                          "yellow": [statusdict['monitor'].get('space') * 0.8,
                                                                     statusdict['monitor'].get('space') * 0.9],
                                                          "red": [statusdict['monitor'].get('space') * 0.9,
                                                                  statusdict['monitor'].get('space')]}},
                                        value=statusdict['monitor'].get('used'),
                                        label='Base',
                                        max=statusdict['monitor'].get('space'),
                                        min=0,
                                    )
                                      ])
                              ),
                              html.Div(
                                  className="mbox mo",
                                  children=html.Div([
                                      daq.Gauge(
                                          id='archive-gauge',
                                          size=120,
                                          color={"gradient": True,
                                                 "ranges": {"green": [0, statusdict['archive'].get('space') * 0.8],
                                                            "yellow": [statusdict['archive'].get('space') * 0.8,
                                                                       statusdict['archive'].get('space') * 0.9],
                                                            "red": [statusdict['archive'].get('space') * 0.9,
                                                                    statusdict['archive'].get('space')]}},
                                          value=statusdict['archive'].get('used'),
                                          label='Archive',
                                          max=statusdict['archive'].get('space'),
                                          min=0,
                                      )
                                      ])
                              ),
                              html.Div(
                                  className='mbox mp',
                                  children=html.Div([
                                      html.Div([dcc.Dropdown(dataoptions, datavalue, id='data-dropdown'),
                                                dcc.Dropdown(keyoptions, keyvalue, id='key-dropdown')], style={"width": "50%"}),
                                      dcc.Graph(id='live-graph', animate=False),
                                      dcc.Interval(
                                          id='graph-update',
                                          interval=5 * 1000,  # in milliseconds
                                          n_intervals=0
                                      )
                                  ])
                              ),
                              html.Div(
                                  className="mbox mq",
                                  children=html.Div([
                                      dash_table.DataTable(crontable,
                                                           columns=[{'id': c, 'name': c} for c in croncols],
                                                           sort_action='native',
                                                           fixed_rows={'headers': True},
                                                           style_cell={'minWidth': 95, 'width': 95, 'maxWidth': 95, 'padding': '5px'},
                                                           style_table={'height': 300},  # default is 500
                                                           style_cell_conditional=[
                                                               {
                                                                   'if': {'column_id': c},
                                                                   'textAlign': 'left'
                                                               } for c in ['job','enabled']
                                                           ],
                                                           style_data={
                                                               'whiteSpace': 'normal',
                                                               'color': 'rgb(230, 230, 250)',
                                                               'fontSize': '14px',
                                                               'backgroundColor': 'rgb(112, 128, 144)'
                                                           },
                                                           style_data_conditional=[
                                                               {
                                                                   'if': {'row_index': 'odd'},
                                                                   'backgroundColor': 'rgb(119, 149, 163)',
                                                               },
                                                               {
                                                                   'if': {
                                                                       'filter_query': '{enabled} = False',
                                                                       'column_id': 'enabled'
                                                                   },
                                                                   'backgroundColor': '#eaf044',
                                                                   'color': 'black'
                                                               },
                                                           ],
                                                           style_header={
                                                               'backgroundColor': '#4D4D4D',
                                                               'color': 'rgb(230, 230, 250)',
                                                               'fontWeight': 'bold',
                                                               'fontSize': '16px',
                                                               'border': '1px solid black'
                                                           },
                                                           ),

                                  ])
                              ),
                             html.Div(
                                 className="mbox mr",
                                 children=html.Div([
                                     dcc.Dropdown(['Sensors', 'Data records'], 'Sensors', id='sensor-dropdown' , style={"width": "50%"}),
                                     dash_table.DataTable(data=dtable,
                                                          id='sensors-table',
                                                          sort_action='native',
                                                          columns=[{'id': c, 'name': c} for c in dcols],
                                                          fixed_rows={'headers': True},
                                                          style_cell={'minWidth': 95, 'width': 95, 'maxWidth': 95, 'padding': '5px'},
                                                          style_table={'height': 150},  # default is 500
                                                          style_cell_conditional=[
                                                                {
                                                                    'if': {'column_id': c},
                                                                    'textAlign': 'left'
                                                                } for c in ['DataID', 'SensorID', 'Actual']
                                                          ],
                                                          style_data={
                                                              'color': 'rgb(230, 230, 250)',
                                                              'fontSize': '14px',
                                                              'backgroundColor': 'rgb(112, 128, 144)'
                                                          },
                                                          style_data_conditional=[
                                                              {
                                                                  'if': {'row_index': 'odd'},
                                                                  'backgroundColor': 'rgb(119, 149, 163)',
                                                              },
                                                              {
                                                                  'if': {
                                                                      'filter_query': '{Data table} = False',
                                                                      'column_id': 'Data table'
                                                                  },
                                                                  'backgroundColor': '#eaf044',
                                                                  'color': 'black'
                                                              },
                                                              {
                                                                  'if': {
                                                                      'filter_query': '{Actual} < 1',
                                                                      'column_id': 'Actual'
                                                                  },
                                                                  'backgroundColor': 'tomato'
                                                              },
                                                              {
                                                                  'if': {
                                                                      'filter_query': '{in DATAINFO} = False',
                                                                      'column_id': 'in DATAINFO'
                                                                  },
                                                                  'backgroundColor': 'tomato'
                                                              },
                                                          ],
                                                          style_header={
                                                              'backgroundColor': '#4D4D4D',
                                                              'color': 'rgb(230, 230, 250)',
                                                              'fontWeight': 'bold',
                                                              'fontSize': '16px',
                                                              'border': '1px solid black'
                                                          }
                                                          ),
                                 ])
                             ),
                             html.Div(
                                 className="mbox mt",
                                 children=html.Div([
                                     html.B('Configuration'),
                                     html.Div(id='live-update-config')
                                 ])
                             ),
                             html.Div(
                                 className="mbox mu",
                                 children=html.Div([
                                     html.B('MARCOS (MagPys automatic real-time collection and organization system)'),
                                     html.P('Version {}'.format(__version__)),
                                     html.P('written by R. Leonhardt, R. Bailey, R. Mandl'),
                                     html.P(['MARTAS on GitHUB:', html.A('click here', href='https://github.com/geomagpy/MARTAS')]),
                                     html.P(['MagPY on GitHUB:',
                                                 html.A('click here', href='https://github.com/geomagpy/geomagpy')]),
                                 ])
                             ),
                 ]
            )
)


@callback(Output('live-update-config', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_config(n):
    #print ("updating config")
    global statusdict
    global archivepath
    global configpath
    global logpath
    global dbcred
    style = {'padding': '5px', 'fontSize': '16px'}
    crontable, croncols, marcoslist, statusdict = get_cron_jobs(statusdict=statusdict, cred=dbcred, archivepath=archivepath,
                                                     logpath=logpath, debug=False)
    pos = {}
    htmllist = []
    cfg = mm.get_conf(configpath)
    #print (cfg)
    htmllist.append(html.P(['Amount of MARCOS jobs: {}'.format(len(marcoslist))]))
    htmllist.append(html.P(['Archivepath: {}'.format(archivepath)]))
    htmllist.append(html.P(['Download jobs: ']))
    return htmllist


@callback(Output('live-update-archivestatus', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_archive_status(n):
    #print ("updating archive stats")
    htmllist = []
    global statusdict
    #global archivepath
    #global logpath
    #global dbcred
    #crontable, croncols, marcoslist, statusdict = get_cron_jobs(statusdict=statusdict, cred=dbcred, archivepath=archivepath, logpath=logpath, debug=False)
    oncolor = '#008003'
    offcolor = '#FF5E5E'
    acolor = offcolor
    ecolor = offcolor
    lcolor = offcolor
    if statusdict['archive'].get('active'):
        acolor = oncolor
    if statusdict['archive'].get('cronenabled'):
        ecolor = oncolor
    if statusdict['archive'].get('logstatus'):
        lcolor = oncolor
    htmllist.append(html.Div([
        daq.Indicator(id='archiveactive', value=True, color=acolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("archive active"),
    ]))
    htmllist.append(html.Div([
        daq.Indicator(id='archiveenabled', value=True, color=ecolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("cron enabled"),
    ]))
    htmllist.append(html.Div([
        daq.Indicator(id='archivelog', value=True, color=lcolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("log status"),
    ]))
    return htmllist


@callback(Output('live-update-monitorstatus', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_monitor_status(n):
    #print ("updating monitor stats")
    htmllist = []
    global statusdict
    #global archivepath
    #global logpath
    #global dbcred
    #crontable, croncols, marcoslist, statusdict = get_cron_jobs(statusdict=statusdict, cred=dbcred, archivepath=archivepath, logpath=logpath, debug=False)
    oncolor = '#008003'
    offcolor = '#FF5E5E'
    acolor = offcolor
    ecolor = offcolor
    lcolor = offcolor
    if statusdict['monitor'].get('active'):
        acolor = oncolor
    if statusdict['monitor'].get('cronenabled'):
        ecolor = oncolor
    if statusdict['monitor'].get('logstatus'):
        lcolor = oncolor
    htmllist.append(html.Div([
        daq.Indicator(id='monitoractive', value=True, color=acolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("monitor active"),
    ]))
    htmllist.append(html.Div([
        daq.Indicator(id='monitorenabled', value=True, color=ecolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("cron enabled"),
    ]))
    htmllist.append(html.Div([
        daq.Indicator(id='monitorlog', value=True, color=lcolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("log status"),
    ]))
    return htmllist


@callback(Output('live-update-dbstatus', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_db_status(n):
    #print ("updating db stats")
    htmllist = []
    global statusdict
    #global archivepath
    #global logpath
    #global dbcred
    #crontable, croncols, marcoslist, statusdict = get_cron_jobs(statusdict=statusdict, cred=dbcred, archivepath=archivepath, logpath=logpath, debug=False)
    oncolor = '#008003'
    offcolor = '#FF5E5E'
    acolor = offcolor
    ecolor = offcolor
    lcolor = offcolor
    if statusdict['database'].get('active'):
        acolor = oncolor
    if statusdict['database'].get('cronenabled'):
        ecolor = oncolor
    if statusdict['database'].get('logstatus'):
        lcolor = oncolor
    htmllist.append(html.Div([
        daq.Indicator(id='dbactive', value=True, color=acolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("database active"),
    ]))
    htmllist.append(html.Div([
        daq.Indicator(id='dbenabled', value=True, color=ecolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("cron enabled"),
    ]))
    htmllist.append(html.Div([
        daq.Indicator(id='dblog', value=True, color=lcolor,
                      style={"float": "left", "margin-right": "10px"}),
        html.P("log status"),
    ]))
    return htmllist


@callback(Output('live-update-marcos1', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos1_status(n):
    #print ("updating marcos 1")
    htmllist=[]
    global statusdict
    if len(marcoslist) > 0:
        m = marcoslist[0]
        htmllist = get_marcos_html(statusdict,m, 1)
    else:
        htmllist.append(html.P(['empty MARCOS slot']))
    return htmllist

@callback(Output('live-update-marcos2', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos2_status(n):
    #print ("updating marcos 2")
    htmllist=[]
    global statusdict
    #crontable, croncols, marcoslist, statusdict = get_cron_jobs(statusdict=statusdict, cred=dbcred, archivepath=archivepath,
    #                                                logpath=logpath, debug=False)
    if len(marcoslist) > 1:
        m = marcoslist[1]
        htmllist = get_marcos_html(statusdict,m, 2)
    else:
        htmllist.append(html.P(['empty slot']))
    return htmllist

@callback(Output('live-update-marcos3', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos3_status(n):
    #print ("updating marcos 3")
    htmllist=[]
    global statusdict
    if len(marcoslist) > 2:
        m = marcoslist[2]
        htmllist = get_marcos_html(statusdict,m, 3)
    else:
        htmllist.append(html.P(['empty slot']))
    return htmllist

@callback(Output('live-update-marcos4', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos4_status(n):
    #print ("updating marcos 4")
    htmllist=[]
    global statusdict
    if len(marcoslist) > 3:
        m = marcoslist[3]
        htmllist = get_marcos_html(statusdict,m, 4)
    else:
        htmllist.append(html.P(['empty slot']))
    return htmllist

@callback(Output('live-update-marcos5', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos5_status(n):
    #print ("updating marcos 5")
    htmllist=[]
    global statusdict
    if len(marcoslist) > 4:
        m = marcoslist[4]
        htmllist = get_marcos_html(statusdict,m, 5)
    else:
        htmllist.append(html.P(['empty slot']))
    return htmllist

@callback(Output('live-update-marcos6', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos6_status(n):
    #print ("updating marcos 6")
    htmllist=[]
    global statusdict
    if len(marcoslist) > 5:
        m = marcoslist[5]
        htmllist = get_marcos_html(statusdict,m, 6)
    else:
        htmllist.append(html.P(['empty slot']))
    return htmllist

@callback(Output('live-update-marcos7', 'children'),
              Input('gauge-update', 'n_intervals'))
def update_marcos7_status(n):
    #print ("updating marcos 7")
    htmllist=[]
    global statusdict
    if len(marcoslist) > 6:
        m = marcoslist[6]
        htmllist = get_marcos_html(statusdict,m, 7)
    else:
        htmllist.append(html.P(['empty slot']))
    return htmllist


@callback(Output('key-dropdown', 'options'),
              Output('key-dropdown', 'value'),
              Input('data-dropdown', 'value'))
def update_keydrop(datavalue):
    result = get_datainfo_from_db(cred=dbcred)
    keyoptions, keyvalue = get_graph_keys(datavalue, result)
    return keyoptions, keyvalue


@callback(Output('sensors-table', 'data'),Output('sensors-table', 'columns'),
              Input('sensor-dropdown', 'value'))
def update_table_update(datavalue):
    #print ("updating table")
    global dbcred
    result = get_datainfo_from_db(cred=dbcred)
    if datavalue.startswith("Sensor"):
        dtable, dcols = convert_datainfo_to_datatable(result)
    else:
        dtable, dcols = convert_datainfo_to_idtable(result)
    cols = [{'id': c, 'name': c} for c in dcols]
    return dtable, cols


@callback(
    Output('live-graph', 'figure'),
    Input('graph-update', 'n_intervals'),
    Input('heightslider', 'value'),
    Input('durationslider', 'value'),
    Input('data-dropdown', 'value'),
    Input('key-dropdown', 'value'),
    prevent_initial_call=True
)
def update_graph(n, hvalue, duration, datavalue, keyvalue):
    #print ("updating graph")
    # read data
    #print ("Get available data sets")
    global dbcred
    result = get_datainfo_from_db(cred=dbcred)

    data, names = get_data(datavalue, [keyvalue], datainfo=result, duration=duration, cred=dbcred)
    fig = make_subplots(rows=len(names), cols=1, vertical_spacing=0.1)
    fig.update_layout(
        plot_bgcolor='#4D4D4D',
        paper_bgcolor='#EDE9E8',
    )
    fig.update_xaxes(
        mirror=True,
        ticks='outside',
        showline=True,
        linecolor='black',
        gridcolor='lightgrey'
    )
    fig.update_yaxes(
        mirror=True,
        ticks='outside',
        showline=True,
        linecolor='black',
        gridcolor='lightgrey'
    )
    fig['layout']['margin'] = {
        'l': 20, 'r': 10, 'b': 10, 't': 10
    }
    fig['layout']['legend'] = {'x': 0, 'y': 1, 'xanchor': 'left'}
    fig.update_layout(height=int(hvalue))

    i = 0
    for name in names:
            i += 1
            fig.add_trace({
                'x': data['time'],
                'y': data[name],
                'name': name,
                'mode': 'lines+markers',
                'type': 'scatter'
            }, i, 1)

    return fig

#if __name__ == '__main__':
#    app.run(host="0.0.0.0", debug=False)




# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timezone
#from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm
import serial
import subprocess
import sys
import time

## Aktive Arduino protocol
## -----------------------

class ActiveArduinoProtocol(object):
    """
    Protocol to read Arduino data (usually from ttyACM0)
    Tested so far only for Arduino Uno on a Linux machine
    The protocol works only if the serial output follows the MagPy convention:
    Up to 99 Sensors are supported identified by unique sensor names and ID's.

    ARDUINO OUTPUT:
        - serial output on ttyACM0 needs to follow the MagPy definition:
            Three data sequences are supported:
            1.) The meta information
                The meta information line contains all information for a specific sensor.
                If more than one sensor is connected, then several meta information
                lines should be sent (e.g. M1:..., M2:..., M99:...)
                Meta lines should be resent once in a while (e.g. every 10-100 data points)
                Example:
                     M1: SensorName: MySensor, SensorID: 12345, SensorRevision: 0001
            2.) The header line
                The header line contains information on the provided data for each sensor.
                The typical format includes the MagPy key, the actual Variable and the unit.
                Key and Variable are separated by an underscore, unit is provided in brackets.
                Like the Meta information the header should be sent out once in a while
                Example:
                     H1: f_F [nT], t1_Temp [degC], var1_Quality [None], var2_Pressure [mbar]
            3.) The data line:
                The data line contains all data from a specific sensor
                Example:
                     D1: 46543.7898, 6.9, 10, 978.000

         - recording starts after meta and header information have been received

    """

    ## need a reference to our WS-MCU gateway factory to dispatch PubSub events
    ##
    def __init__(self, client, sensordict, confdict):
        self.client = client #self.wsMcuFactory = wsMcuFactory
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.counter = {}
        self.initserial = True

        # Commands (take them sensordesc)
        # i.e. sensordesc : owT-swD
        commandlist = sensordict.get('sensordesc').strip().replace('\n','').split('-')
        if not 'arduino sensors' in commandlist:
            cdict = {}
            for idx,com in enumerate(commandlist):
                cdict['data{}'.format(idx)] = com
            self.commands = [cdict]
            print ("Commands from sensordesc:", self.commands)
        else:
            self.commands = [{'data1':'owT','data2':'swD'}]
            print ("Default commands selected")
        self.hexcoding = False
        self.eol = '\r\n'

        # Serial communication data (from confdict)
        self.sensor = sensordict.get('sensorid')
        self.baudrate=int(sensordict.get('baudrate'))
        self.port = confdict['serialport']+sensordict.get('port')
        self.parity=sensordict.get('parity')
        self.bytesize=sensordict.get('bytesize')
        self.stopbits=sensordict.get('stopbits')
        self.timeout=2 # should be rate depended

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # switch on debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # Arduino specific
        self.board = self.sensor
        if self.debug:
            log.msg("DEBUG - Running on board {}".format(self.board))
        # get existing sensors for the relevant board
        self.existinglist = mm.get_sensors(confdict.get('sensorsconf'),identifier='?',secondidentifier=self.board)
        if self.debug:
            log.msg("DEBUG - Arduino sensors: {}".format(self.existinglist))
        self.sensor = ''
        self.ser = None

        # none is verified when initializing
        self.verifiedids = []
        self.infolist = []
        self.headlist = []

    def datetime2array(self,t):
        return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

    def open_connection(self):
        log.msg("connecting to serial port")
        ser = serial.Serial(self.port, baudrate=int(self.baudrate), parity=self.parity, bytesize=int(self.bytesize), stopbits=int(self.stopbits), timeout=self.timeout)
        # wait for arduino (tested: 2 sec are necessary)
        time.sleep(2)
        self.ser = ser
        return ser

    def restart(self):
        try:
            subprocess.check_call(['/home/debian/.martas/runmartas.sh', 'restart'])
        except subprocess.CalledProcessError:
            log.msg('SerialCall: check_call didnt work')
            pass # handle errors in the called executable
        except:
            log.msg('SerialCall: check call problem')
            pass # executable not found
        log.msg('SerialCall: Restarted martas process')


    def processArduinoData(self, sensorid, meta, data):
        """Convert raw ADC counts into SI units as per datasheets"""
        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        #actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        #outtime = datetime.strftime(currenttime, "%H:%M:%S")
        #timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        filename = outdate

        datearray = self.datetime2array(currenttime)

        #datearray = mm.time_to_array(timestamp)
        packcode = '6hL'
        #sensorid = self.sensordict.get(idnum)
        #events = self.eventdict.get(idnum).replace('evt','').split(',')[3:-1]

        values = []
        multiplier = []
        for dat in data:
            try:
                values.append(float(dat))
                datearray.append(int(float(dat)*10000))
                packcode = packcode + 'l'
                multiplier.append(10000)
            except:
                log.msg('{} protocol: Error while appending data to file (non-float?): {}'.format(self.sensordict.get('protocol'),dat) )

        try:
            data_bin = struct.pack('<'+packcode,*datearray) #little endian
        except:
            log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))
            pass

        key = '['+str(meta.get('SensorKeys')).replace("'","").strip()+']'
        ele = '['+str(meta.get('SensorElements')).replace("'","").strip()+']'
        unit = '['+str(meta.get('SensorUnits')).replace("'","").strip()+']'
        multplier = str(multiplier).replace(" ","")
        # Correct some common old problem
        unit = unit.replace('deg C', 'degC')

        header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, key, ele, unit, multplier, packcode, struct.calcsize('<'+packcode))

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)

        return ','.join(list(map(str,datearray))), header


    def analyzeHeader(self, line):
        log.msg("  -> Received unverified header data")
        if self.debug:
            log.msg("DEBUG -> Header looks like: {}".format(line))
        headdict = {}

        head = line.strip().split(':')
        headernum = int(head[0].strip('H'))
        header = head[1].split(',')

        try:
            varlist = []
            keylist = []
            unitlist = []
            for elem in header:
                an = elem.strip(']').split('[')
                try:
                    if len(an) < 1:
                        log.err("Arduino: error when analyzing header")
                        return {}
                except:
                    log.err("Arduino: error when analyzing header")
                    return {}
                var = an[0].split('_')
                key = var[0].strip().lower()
                variable = var[1].strip()
                unit = an[1].strip()
                keylist.append(key)
                varlist.append(variable)
                unitlist.append(unit)
            headdict['ID'] = headernum
            headdict['SensorKeys'] = ','.join(keylist)
            headdict['SensorElements'] = ','.join(varlist)
            headdict['SensorUnits'] = ','.join(unitlist)
        except:
            # in case an incomplete header is available
            # will be read next time completely
            return {}

        return headdict


    def getSensorInfo(self, line):
        log.msg("  -> Received unverified sensor information")
        if self.debug:
            log.msg("DEBUG -> Sensor information line looks like: {}".format(line))
        metadict = {}
        try:
            metaident = line.strip().split(':')
            metanum = int(metaident[0].strip('M'))
            meta = line[3:].strip().split(',')
            for elem in meta:
                el = elem.split(':')
                metadict[el[0].strip()] = el[1].strip()
        except:
            log.err('  -> Could not interprete sensor information - skipping')
            return {}

        if not 'SensorRevsion' in metadict:
            metadict['SensorRevision'] = '0001' # dummy value
        if not 'SensorID' in metadict:
            metadict['SensorID'] = '12345' # dummy value
        if not 'SensorName' in metadict:
            print("No Sensorname provided - aborting")
            return {}
        metadict['ID'] = metanum
        return metadict


    def GetArduinoSensorList(self, line):
        """
         DESCRIPTION:
             Will analysze data line and a list of existing IDs.
             Please note: when initializing, then existing data will be taken from
             Arduino Block.
             - If ID is existing, this method will return its ID, sensorid, meta info
               as used by process data, and data.
             - If ID not yet existing: line will be scanned until all meta info is
               available. Method will return ID 0 and empty fields.
             - If meta info is coming: sensorids will be quickly checked against existing.
               If not existing and not yet used - ID will be added to sensors.cfg and existing
               If not existing but used so far - ID in file is wrong -> warning
               If existing and ID check OK: continue
        TODO: path number might change if a sensor is disconnected
              problem with sensorID and ID number...

        PARAMETER:
            existinglist: [list] [[1,2,...],['BM35_xxx_0001','SH75_xxx_0001',...]]
                          idnum is stored in sensordict['path'] (like ow)
        """
        #existingpathlist = [line.get('path') for line in existinglist]
        if self.debug:
            log.msg("DEBUG - running get Sensorlist")

        evdict = {}
        meta = 'not known'
        data = '1.0'

        lineident = line.split(':')
        try:
            idnum = int(lineident[0][1:])
        except:
            idnum = 999
        if not idnum == 999:
            if not idnum in self.verifiedids:
                if line.startswith('H'):
                    # analyse header
                    headdict = self.analyzeHeader(line)
                    self.headlist.append(headdict)
                elif line.startswith('M'):
                    # analyse sensorinformation
                    infodict = self.getSensorInfo(line)
                    self.infolist.append(infodict)
                    # add values to metadict
                # at this stage i have the correct sensorid in infodict
                if idnum in [idict.get('ID') for idict in self.infolist] and idnum in [hdict.get('ID') for hdict in self.headlist]:
                    # idnums found in line, present in head and meta data
                    sensoridenti = [idict.get('SensorName') for idict in self.infolist if str(idict.get('ID')) == str(idnum)]
                    # get sensorid as well for the identificationnumner
                    sensoridininfo = [idict.get('SensorID') for idict in self.infolist if str(idict.get('ID')) == str(idnum)]
                    # board is already selected
                    seldict2 = [edict for edict in self.existinglist if str(edict.get('path')) == str(idnum)]
                    # if a single sensor id is found:
                    # it might happen that idnum (path in existinglist) is not referring
                    # to the correct sensorid. This might happen if e.g. one ow temperature
                    # sensor is disconnected, when previously two have been attached.
                    # after restart the remaining sensor will have the lowest id, not
                    # necessarily matching the originally stored one.
                    # Therefor we check it and eventually assign a new id/path
                    if len(sensoridininfo) == 1:
                        sensid = sensoridininfo[0]
                        seldict1 = [edict for edict in self.existinglist if edict.get('sensorid').find(sensid) > -1]
                    else:
                        seldict1 = seldict2
                    if not seldict1==seldict2:
                        # sensorid not matching the sensoid stored with this idnum/path
                        seldict2 = seldict1
                        # therefore update pathid in selfexistinglist
                        self.existinglist = [e for e in self.existinglist if not str(e.get('path')) == str(idnum)]
                        seldicttmp = seldict1[-1]
                        seldicttmp['path'] = str(idnum)
                        self.existinglist.append(seldicttmp)
                    seldict3 = [edict for edict in seldict2 if str(edict.get('name')) == str(sensoridenti[0])]
                    # Select correct idnum, so that sesnorid's are fitting
                    if not len(seldict3) > 0 and len(sensoridenti) > 0:
                        log.msg("Arduino: Sensor {} not yet existing -> adding to existinglist".format(sensoridenti[0]))
                        relevantdict = [idict for idict in self.infolist if str(idict.get('ID')) == str(idnum)][0]
                        values = {}
                        values['path'] = idnum
                        values['serialnumber'] = relevantdict.get('SensorID')
                        values['name'] = relevantdict.get('SensorName')
                        values['protocol'] = 'Arduino'
                        values['port'] = self.board
                        values['ptime'] = relevantdict.get('DataTimeProtocol','-')
                        values['pierid'] = relevantdict.get('DataPier','-')
                        values['sensordesc'] = relevantdict.get('SensorDecription','-')
                        values['sensorgroup'] = relevantdict.get('SensorGroup','-')
                        values['revision'] = relevantdict.get('SensorRevision')
                        values['stack'] = 0
                        values['sensorid'] = sensoridenti[0]+'_'+relevantdict.get('SensorID')+'_'+relevantdict.get('SensorRevision')
                        log.msg("Arduino: Writing new sensor input to sensors.cfg ...")
                        success = mm.add_sensors(self.confdict.get('sensorsconf'), values, block='Arduino')
                        #success = mm.add_sensors(self.confdict.get('sensorsconf'), values, block='Arduino')
                        print ("Add sensor done")
                        self.existinglist.append(values)
                        print ("sensor added to arduino list")

                    elif len(seldict3) > 0:
                        log.msg("Arduino: Sensor {} identified and verified".format(sensoridenti[0]))
                        self.verifiedids.append(idnum)
                    else:
                        log.err('Apparently a sensor is connected which does not correspond to the provided information in sensors.cfg - Please clarify (either delete sensor.cfg input for {} or check your arduino code'.format(sensoridenti[0]))

            else:
                if line.startswith('D'):
                    dataar = line.strip().split(':')
                    dataident = int(dataar[0].strip('D'))
                    meta = [headdict for headdict in self.headlist if str(headdict.get('ID')) == str(dataident)][0]
                    evdict = [edict for edict in self.existinglist if str(edict.get('path')) == str(idnum)][0]
                    data = dataar[1].strip().split(',')

        else:
            # invalid return value found
            if self.debug:
                log.msg("DEBUG - Invalid return value found: {}".format(line))

        return evdict, meta, data


    def sendRequest(self):

        # connect to serial
        ser = self.ser
        if not ser or not ser.isOpen():
            ser = self.open_connection()
        #ser = serial.Serial(self.port, baudrate=int(self.baudrate), parity=self.parity, bytesize=int(self.bytesize), stopbits=int(self.stopbits), timeout=self.timeout)
        # wait for arduino (tested: 2 sec are necessary)
        #time.sleep(2)

        def write_read(ser,command,end="MARTASEND",maxcnt=20,debug=False):
            #data = b""
            dedata = ""
            all = ""
            cnt=0
            ser.flush()
            try:
                ser.write(command)
                #responsetime1 = datetime.now(timezone.utc).replace(tzinfo=None)
                time.sleep(.1)
                while not dedata.find(end) > -1 and not cnt == maxcnt:
                   cnt+=1
                   data = ser.readline()
                   #responsetime2 = datetime.now(timezone.utc).replace(tzinfo=None)
                   if sys.version_info >= (3, 0):
                       dedata = str(data.decode("utf-8"))
                   else:
                       dedata = str(data)
                   if not dedata:
                       if debug:
                           log.msg("DEBUG - got empty line")
                   else:
                       all += dedata
                # time estimate: mean(resp1,resp2)-0.1sec
                return all
            except serial.SerialException as e:
                if debug:
                    log.msg("SerialException found ({})".format(e)) # - restarting martas ...")
                time.sleep(0.1)
                #self.restart()
            except:
                log.msg("Other exception found")
                raise

            if cnt == maxcnt:
                if debug:
                    log.msg('Maximum count {} was reached without MARTASEND'.format(maxcnt))
            return all


        def get_commands(condition="data"):
            # create a command list from given values in sensors.cfg
            # also check for python2 and py3
            coms=[]
            for sensordict in self.commands:
                for item in sensordict:
                    command = sensordict.get(item)
                    n = command.count(":")
                    miss=-(n-2)
                    for n in range(miss):
                        command += ':'
                    command = command+self.eol
                    if sys.version_info >= (3, 0):
                        command = command.encode('utf-8')
                    if item.startswith(condition):
                        coms.append(command)
            return coms

        def run_analysis(answer):
            lines = answer.split("\n")
            for line in lines:
                try:
                    if line and len(line)>2 and (line[2] == ':' or line[3] == ':'):
                        self.analyzeline(line)
                except:
                    pass

        coms = get_commands()
        if self.debug:
            log.msg("DEBUG - sending the following commands to serial port: {}".format(coms))

        if ser.isOpen():
            for command in coms:
                answer = write_read(ser,command,debug=self.debug)
                if self.debug:
                    log.msg("DEBUG - received the following answer for command {}: {}".format(command,answer))
                run_analysis(answer)
            #ser.close()
        else:
            print ("lib activearduino: Could not open serial port")

    def analyzeline(self, line):

        if self.debug:
            log.msg("Received line: {}".format(line))

        # extract only ascii characters
        line = ''.join(filter(lambda x: x in string.printable, str(line)))

        # Create a list of sensors like for OW
        # dispatch with the appropriate sensor
        evdict, meta, data = self.GetArduinoSensorList(line)

        if len(evdict) > 0:
            sensorid = evdict.get('name')+'_'+evdict.get('serialnumber')+'_'+evdict.get('revision')
            topic = self.confdict.get('station') + '/' + sensorid
            pdata, head = self.processArduinoData(sensorid, meta, data)

            senddata = False
            try:
                coll = int(evdict.get('stack'))
            except:
                coll = 0

            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(pdata)
                    self.datacnt += 1
                else:
                    senddata = True
                    pdata = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", pdata, qos=self.qos)
                # If multiple sensors are connected, self.count needs to be a dict (with topic)
                # Initialize counter for each topic
                if self.counter.get(sensorid,'') == '':
                    self.counter[sensorid] = 0
                cnt = self.counter.get(sensorid)

                if cnt == 0:
                    ## 'Add' is a string containing dict info like:
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( evdict.get('sensorid',''),self.confdict.get('station','').strip(),evdict.get('pierid','').strip(),evdict.get('protocol','').strip(),evdict.get('sensorgroup','').strip(),evdict.get('sensordesc','').strip(),evdict.get('ptime','').strip() )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)

                cnt += 1
                #self.count += 1
                if cnt >= self.metacnt:
                    cnt = 0

                # update counter in dict
                self.counter[sensorid] = cnt
